"""Views for the diff viewer."""

from __future__ import annotations

import logging
import os
import re
import traceback
from io import BytesIO
from typing import Any, Mapping, Optional, TYPE_CHECKING, Union
from zipfile import ZipFile

from django.conf import settings
from django.core.paginator import InvalidPage, Paginator
from django.http import (HttpResponse,
                         HttpResponseNotFound,
                         HttpResponseNotModified,
                         HttpResponseServerError,
                         Http404)
from django.shortcuts import get_object_or_404
from django.template.loader import render_to_string
from django.urls import NoReverseMatch
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django.views.generic.base import TemplateView, View
from djblets.util.http import encode_etag, etag_if_none_match, set_etag
from housekeeping import deprecate_non_keyword_only_args
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name
from typing_extensions import NotRequired, TypedDict

from reviewboard.deprecation import (
    RemovedInReviewBoard90Warning,
    RemovedInReviewBoard10_0Warning,
)
from reviewboard.diffviewer.commit_utils import (
    SerializedCommitHistoryDiffEntry,
    diff_histories)
from reviewboard.diffviewer.diffutils import get_diff_files
from reviewboard.diffviewer.errors import PatchError, UserVisibleError
from reviewboard.diffviewer.models import DiffCommit, DiffSet, FileDiff
from reviewboard.diffviewer.renderers import (
    DiffRenderer,
    get_diff_renderer,
    get_diff_renderer_class,
)
from reviewboard.diffviewer.settings import DiffSettings
from reviewboard.scmtools.errors import FileNotFoundError
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.http import HttpRequest

    from reviewboard.diffviewer.diffutils import SerializedDiffFile
    from reviewboard.diffviewer.models.diffcommit import SerializedDiffCommit


logger = logging.getLogger(__name__)


class SerializedPaginationInfo(TypedDict):
    """Serialized information about pagination.

    Version Added:
        7.0
    """

    #: Whether the diff has been paginated.
    is_paginated: bool

    #: The current page number.
    current_page: int

    #: The total number of pages.
    pages: int

    #: The set of all page numbers.
    page_numbers: list[int]

    #: Whether there's another page after the current one.
    has_next: bool

    #: Whether there's another page before the current one.
    has_previous: bool

    #: The number of the next page, if available.
    next_page: NotRequired[int]

    #: The number of the previous page, if available.
    previous_page: NotRequired[int]


class SerializedRevisionInfo(TypedDict):
    """Serialized information about diff revisions.

    Version Added:
        7.0
    """

    #: The current diff revision.
    revision: int

    #: Whether the current page is an interdiff.
    is_interdiff: bool

    #: The current interdiff revision, if available.
    interdiff_revision: Optional[int]

    #: The current base commit ID, when viewing a commit range.
    base_commit_id: Optional[int]

    #: The current tip commit ID, when viewing a commit range.
    tip_commit_id: Optional[int]

    #: The most recent diff revision, if available.
    latest_revision: NotRequired[Optional[int]]

    #: Whether the current page is showing a draft diff.
    is_draft_diff: NotRequired[bool]

    #: Whether the current page is showing a draft interdiff.
    is_draft_interdiff: NotRequired[bool]


class SerializedDiffContext(TypedDict):
    """Serialized diff information.

    Version Added:
        7.0
    """

    #: The set of commits for the current diff revision.
    commits: Optional[list[SerializedDiffCommit]]

    #: The diff of commits between two diff revisions.
    #:
    #: This will only be populated when viewing an interdiff.
    commit_history_diff: Optional[list[SerializedCommitHistoryDiffEntry]]

    #: A list of filename patterns to limit the view to.
    filename_patterns: list[str]

    #: Pagination information for the current diff.
    pagination: SerializedPaginationInfo

    #: Revision information for the current diff.
    revision: SerializedRevisionInfo


class DiffViewerContext(TypedDict):
    """Render context for the diff viewer views.

    Version Added:
        7.0
    """

    #: Whether to collapse all expandable diff context.
    collapseall: bool

    #: The template context information for the diff.
    diff_context: SerializedDiffContext

    #: The current diff.
    diffset: DiffSet

    #: The current interdiff, if available.
    interdiffset: Optional[DiffSet]

    #: The list of all files in the current diff/interdiff.
    files: list[SerializedDiffFile]


def get_collapse_diff(request):
    if request.GET.get('expand', False):
        return False
    elif request.GET.get('collapse', False):
        return True
    elif 'collapsediffs' in request.COOKIES:
        return (request.COOKIES['collapsediffs'] == "True")
    else:
        return True


class DiffViewerView(TemplateView):
    """Renders the main diff viewer.

    This renders the diff viewer for a given DiffSet (or an interdiff
    between two DiffSets). It handles loading information on the diffs,
    generating the side-by-side view, and pagination.

    The view expects the following parameters to be provided:

    ``diffset``
        The DiffSet to render.

    The following may also be provided:

    ``interdiffset``
        A DiffSet object representing the other end of an interdiff range.

    The following query parameters can be passed in on the URL:

    ``?expand=1``
        Expands all files within the diff viewer.

    ``?collapse=1``
        Collapses all files within the diff viewer, showing only
        modifications and a few lines of context.

    ``?file=<id>``
        Renders only the FileDiff represented by the provided ID.

    ``?filenames=<pattern>[,<pattern>,...]``
        Renders files matching the given filenames or
        :py:mod:`patterns <fnmatch>`. Patterns are case-sensitive.

    ``?page=<pagenum>``
        Renders diffs found on the given page number, if the diff viewer
        is paginated.

    ``?base-commit-id=<id>``
        The ID of the base commit to use to generate the diff for diffs created
        with multiple commits.

        Only changes from after the specified commit will be included in the
        diff.

    ``?tip-commit-id=<id>``
        The ID of the tip commit to use to generate the diff for diffs created
        created with history.

        No changes from beyond this commit will be included in the diff.
    """

    template_name = 'diffviewer/view_diff.html'
    fragment_error_template_name = 'diffviewer/diff_fragment_error.html'

    def get(
        self,
        request: HttpRequest,
        diffset: DiffSet,
        interdiffset: Optional[DiffSet] = None,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Handle GET requests for this view.

        This will render the full diff viewer based on the provided
        parameters.

        The full rendering time will be logged.

        If there's any exception thrown during rendering, an error page
        with a traceback will be returned instead.
        """
        self.collapse_diffs = get_collapse_diff(request)

        if interdiffset:
            logger.debug('Generating diff viewer page for interdiffset '
                         'ids %s-%s',
                         diffset.pk, interdiffset.pk,
                         extra={'request': request})
        else:
            logger.debug('Generating diff viewer page for filediff id %s',
                         diffset.pk,
                         extra={'request': request})

        try:
            response = super().get(
                request, diffset=diffset, interdiffset=interdiffset,
                *args, **kwargs)

            if interdiffset:
                logger.debug('Done generating diff viewer page for '
                             'interdiffset ids %s-%s',
                             diffset.pk, interdiffset.pk,
                             extra={'request': request})
            else:
                logger.debug('Done generating diff viewer page for filediff '
                             'id %s',
                             diffset.pk,
                             extra={'request': request})

            return response
        except Exception as e:
            if interdiffset:
                interdiffset_id = interdiffset.pk
            else:
                interdiffset_id = None

            logger.exception('%s.get: Error rendering diff for diffset '
                             'ID=%s, interdiffset ID=%s: %s',
                             self.__class__.__name__,
                             diffset.pk,
                             interdiffset_id,
                             e,
                             extra={'request': request})

            return exception_traceback(request, e, self.template_name)

    def render_to_response(self, *args, **kwargs) -> HttpResponse:
        """Render the page to an HttpResponse.

        This renders the diff viewer page, based on the context data
        generated, and sets cookies before returning an HttpResponse to
        the client.

        Args:
            *args (tuple):
                Positional arguments to pass through to the parent class.

            **kwargs (tuple):
                Keyword arguments to pass through to the parent class.

        Returns:
            django.http.HttpResponse:
            The response to send back to the client.
        """
        response = super().render_to_response(*args, **kwargs)
        response.set_cookie('collapsediffs', self.collapse_diffs)

        return response

    @deprecate_non_keyword_only_args(RemovedInReviewBoard90Warning)
    def get_context_data(
        self,
        *,
        diffset: DiffSet,
        interdiffset: Optional[DiffSet],
        all_commits: list[DiffCommit],
        base_commit: Optional[DiffCommit],
        tip_commit: Optional[DiffCommit],
        extra_context: Optional[dict[str, Any]] = None,
        **kwargs,
    ) -> DiffViewerContext:
        """Calculate and return data used for rendering the diff viewer.

        This handles all the hard work of generating the data backing the
        side-by-side diff, handling pagination, and more. The data is
        collected into a context dictionary and returned for rendering.

        Version Changed:
            7.0:
            * Added typed dicts for return data.
            * Made arguments keyword-only.
            * Added ``all_commits``, ``base_commit``, and ``tip_commit``
              arguments.

        Args:
            diffset (reviewboard.diffviewer.models.DiffSet):
                The diffset being viewed.

            interdiffset (reviewboard.diffviewer.models.DiffSet):
                The interdiff diffset, if present.

            all_commits (list of reviewboard.diffviewer.models.DiffCommit):
                The list of all commits in all diffs on the review request.

            base_commit (reviewboard.diffviewer.models.DiffCommit):
                The base commit, when viewing a commit range.

            tip_commit (reviewboard.diffviewer.models.DiffCommit):
                The tip commit, when viewing a commit range.

            extra_context (dict, optional):
                Extra information to add to the context.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            dict:
            Context to use when rendering the template.
        """
        try:
            filename_patterns = \
                re.split(r',+\s*', self.request.GET['filenames'].strip())
        except KeyError:
            filename_patterns = []

        diff_settings = DiffSettings.create(request=self.request)
        files = get_diff_files(
            diffset=diffset,
            interdiffset=interdiffset,
            request=self.request,
            filename_patterns=filename_patterns,
            base_commit=base_commit,
            tip_commit=tip_commit,
            diff_settings=diff_settings)

        paginator = Paginator(files,
                              diff_settings.paginate_by,
                              diff_settings.paginate_orphans)

        page_num = int(self.request.GET.get('page', 1))

        if self.request.GET.get('file', False):
            file_id = int(self.request.GET['file'])

            for i, f in enumerate(files):
                if f['filediff'].pk == file_id:
                    page_num = min(i // paginator.per_page + 1,
                                   paginator.num_pages)

                    break

        try:
            page = paginator.page(page_num)
        except InvalidPage:
            page = paginator.page(paginator.num_pages)

        diff_context: SerializedDiffContext = {
            'commits': None,
            'commit_history_diff': None,
            'filename_patterns': filename_patterns,
            'revision': {
                'revision': diffset.revision,
                'is_interdiff': interdiffset is not None,
                'interdiff_revision': (interdiffset.revision
                                       if interdiffset else None),
                'base_commit_id': base_commit and base_commit.pk,
                'tip_commit_id': tip_commit and tip_commit.pk,
            },
            'pagination': {
                'is_paginated': page.has_other_pages(),
                'current_page': page.number,
                'pages': paginator.num_pages,
                'page_numbers': list(paginator.page_range),
                'has_next': page.has_next(),
                'has_previous': page.has_previous(),
            },
        }

        if page.has_next():
            diff_context['pagination']['next_page'] = page.next_page_number()

        if page.has_previous():
            diff_context['pagination']['previous_page'] = \
                page.previous_page_number()

        if diffset.commit_count > 0:
            diff_commits = [
                commit
                for commit in all_commits
                if commit.diffset_id == diffset.pk
            ]

            diff_context['commits'] = [
                commit.serialize()
                for commit in diff_commits
            ]

            if interdiffset:
                interdiff_commits = [
                    commit
                    for commit in all_commits
                    if commit.diffset_id == interdiffset.pk
                ]

                diff_context['commit_history_diff'] = [
                    entry.serialize()
                    for entry in diff_histories(diff_commits,
                                                interdiff_commits)
                ]

                diff_context['commits'].extend(
                    commit.serialize()
                    for commit in interdiff_commits
                )

        context: DiffViewerContext = {
            'diff_context': diff_context,
            'diffset': diffset,
            'interdiffset': interdiffset,
            'files': list(page.object_list),
            'collapseall': self.collapse_diffs,
        }

        if extra_context is not None:
            context.update(extra_context)

        return context


class DiffFragmentView(View):
    """Renders a fragment from a file in the diff viewer.

    Based on the diffset data and other arguments provided, this will render
    a fragment from a file in a diff. This may be the entire file, or some
    chunk within.

    The view expects the following parameters to be provided:

        * diffset_or_id
          - A DiffSet object or the ID for one.

        * filediff_id
          - The ID of a FileDiff within the DiffSet.

    The following may also be provided:

        * interdiffset_or_id
          - A DiffSet object or the ID for one representing the other end of
            an interdiff range.

        * interfilediff_id
          - A FileDiff ID for the other end of a revision range.

        * chunk_index
          - The index (0-based) of the chunk to render. If left out, the
            entire file will be rendered.

    Both ``filediff_id` and ``interfilediff_id`` need to be available in the
    URL (or otherwise passed to :py:meth:`get`). ``diffset_or_id`` and
    ``interdiffset_or_id`` are needed in :py:meth:`process_diffset_info`, and
    so must be passed either in the URL or in a subclass's definition of
    that method.

    The following query parameters can be passed in on the URL:

    ``?lines-of-context=<count>``
        A number of lines of context to include above and below the chunk.

    ``?base-filediff-id<=id>``
        The primary key of the base FileDiff.

        This parameter is ignored if the review request was created without
        commit history support.

        This conflicts with the ``interfilediff_id``.
    """

    template_name = 'diffviewer/diff_file_fragment.html'
    error_template_name = 'diffviewer/diff_fragment_error.html'
    patch_error_template_name = 'diffviewer/diff_fragment_patch_error.html'

    def get(self, request, *args, **kwargs):
        """Handle GET requests for this view.

        This will create the renderer for the diff fragment, render it, and
        return it.

        If there's an error when rendering the diff fragment, an error page
        will be rendered and returned instead.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            *args (tuple):
                Additional positional arguments for the view.

            **kwargs (dict):
                Additional keyword arguments for the view.

        Returns:
            django.http.HttpResponse:
            A response containing the rendered fragment.
        """
        filediff_id = kwargs.get('filediff_id')
        interfilediff_id = kwargs.get('interfilediff_id')
        chunk_index = kwargs.get('chunk_index')

        base_filediff_id = request.GET.get('base-filediff-id')

        try:
            renderer_settings = self._get_renderer_settings(**kwargs)
            etag = self.make_etag(
                request=request,
                renderer_settings=renderer_settings,
                **kwargs)

            if etag_if_none_match(request, etag):
                return HttpResponseNotModified()

            diff_info = self.process_diffset_info(
                base_filediff_id=base_filediff_id,
                **kwargs)
        except Http404:
            raise
        except Exception as e:
            logger.exception('%s.get: Error when processing diffset info '
                             'for filediff ID=%s, interfilediff ID=%s, '
                             'chunk_index=%s: %s',
                             self.__class__.__name__,
                             filediff_id,
                             interfilediff_id,
                             chunk_index,
                             e,
                             extra={'request': request})

            return exception_traceback(self.request, e,
                                       self.error_template_name)

        kwargs.update(diff_info)

        try:
            context = self.get_context_data(**kwargs)

            renderer = self.create_renderer(
                request=request,
                context=context,
                renderer_settings=renderer_settings,
                **kwargs)
            response = renderer.render_to_response(request)
        except PatchError as e:
            logger.warning(
                '%s.get: PatchError when rendering diffset for filediff '
                'ID=%s, interfilediff ID=%s, chunk_index=%s: %s',
                self.__class__.__name__,
                filediff_id,
                interfilediff_id,
                chunk_index,
                e,
                extra={'request': request})

            try:
                url_kwargs = {
                    key: kwargs[key]
                    for key in ('chunk_index', 'interfilediff_id',
                                'review_request_id', 'filediff_id',
                                'revision', 'interdiff_revision')
                    if key in kwargs and kwargs[key] is not None
                }

                bundle_url = local_site_reverse('patch-error-bundle',
                                                kwargs=url_kwargs,
                                                request=request)
            except NoReverseMatch:
                # We'll sometimes see errors about this failing to resolve when
                # web crawlers start accessing fragment URLs without the proper
                # attributes. Ignore them.
                bundle_url = ''

            if e.rejects:
                lexer = get_lexer_by_name('diff')
                formatter = HtmlFormatter()
                rejects = highlight(e.rejects, lexer, formatter)
            else:
                rejects = None

            return HttpResponseServerError(render_to_string(
                template_name=self.patch_error_template_name,
                context={
                    'bundle_url': bundle_url,
                    'file': diff_info['diff_file'],
                    'filename': os.path.basename(e.filename),
                    'patch_output': e.error_output,
                    'rejects': mark_safe(rejects),
                },
                request=request))
        except FileNotFoundError as e:
            return HttpResponseServerError(render_to_string(
                template_name=self.error_template_name,
                context={
                    'error': e,
                    'file': diff_info['diff_file'],
                },
                request=request))
        except Exception as e:
            logger.exception('%s.get: Error when rendering diffset for '
                             'filediff ID=%s, interfilediff ID=%s, '
                             'chunkindex=%s: %s',
                             self.__class__.__name__,
                             filediff_id,
                             interfilediff_id,
                             chunk_index,
                             e,
                             extra={'request': request})

            return exception_traceback(
                self.request, e, self.error_template_name,
                extra_context={
                    'file': diff_info['diff_file'],
                })

        if response.status_code == 200:
            set_etag(response, etag)

        return response

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def make_etag(
        self,
        *,
        renderer_settings: Mapping[str, Any],
        filediff_id: int,
        interfilediff_id: (int | None) = None,
        request: (HttpRequest | None) = None,
        **kwargs,
    ) -> str:
        """Return an ETag identifying this render.

        Version Changed:
            8.0:
            * Deprecated non-keyword arguments.
            * Added the ``request`` parameter.

        Args:
            renderer_settings (dict):
                The settings determining how to render this diff.

                The following keys are required: ``collapse_all`` and
                ``highlighting``.

                The following key is optional: ``show_deleted``.

            filediff_id (int):
                The ID of the
                :py:class:`~reviewboard.diffviewer.models.filediff.FileDiff`
                being rendered.

            interfilediff_id (int):
                The ID of the
                :py:class:`~reviewboard.diffviewer.models.filediff.FileDiff` on
                the other side of the diff revision, if viewing an interdiff.

            request (django.http.HttpRequest, optional):
                The request from the client.

                Version Added:
                    8.0

            **kwargs (dict):
                Additional keyword arguments passed to the function.

        Returns:
            str:
            The encoded ETag identifying this render.
        """
        return encode_etag(
            '%s:%s:%s:%s:%s:%s'
            % (get_diff_renderer_class(),
               renderer_settings['diff_settings'].state_hash,
               renderer_settings['collapse_all'],
               filediff_id,
               interfilediff_id,
               settings.TEMPLATE_SERIAL))

    def process_diffset_info(
        self,
        diffset_or_id: DiffSet | int,
        filediff_id: int,
        interfilediff_id: (int | None) = None,
        interdiffset_or_id: (DiffSet | int | None) = None,
        base_filediff_id: (int | None) = None,
        **kwargs,
    ) -> Mapping[str, Any]:
        """Process and return information on the desired diff.

        The diff IDs and other data passed to the view can be processed and
        converted into DiffSets. A dictionary with the DiffSet and FileDiff
        information will be returned.

        Args:
            diffset_or_id (reviewboard.diffviewer.models.diffset.DiffSet or
                           int):
                The DiffSet object, or the ID of the diffset.

            filediff_id (int):
                The ID of the FileDiff.

            interfilediff_id (int, optional):
                The ID of the FileDiff for rendering an interdiff, if present.

            interdiffset_or_id (reviewboard.diffviewer.models.diffset.DiffSet
                                or int):
                The diffset object, or the ID of the diffset.

            base_filediff_id (int):
                The ID of the base FileDiff to use, if present. This may only
                be provided if ``filediff_id`` is provided and
                ``interfilediff_id`` is not.

            **kwargs (dict, unused):
                Additional keyword arguments, for future expansion.

        Returns:
            dict:
            Information about the diff.
        """
        # Depending on whether we're invoked from a URL or from a wrapper
        # with precomputed diffsets, we may be working with either IDs or
        # actual objects. If they're objects, just use them as-is. Otherwise,
        # if they're IDs, we want to grab them both (if both are provided)
        # in one go, to save on an SQL query.
        diffset = None
        interdiffset = None

        diffset_ids = []

        if isinstance(diffset_or_id, DiffSet):
            diffset = diffset_or_id
        else:
            diffset_ids.append(diffset_or_id)

        if interdiffset_or_id:
            if isinstance(interdiffset_or_id, DiffSet):
                interdiffset = interdiffset_or_id
            else:
                diffset_ids.append(interdiffset_or_id)

        if diffset_ids:
            diffsets = DiffSet.objects.filter(pk__in=diffset_ids)

            if len(diffsets) != len(diffset_ids):
                raise Http404

            for temp_diffset in diffsets:
                if temp_diffset.pk == diffset_or_id:
                    diffset = temp_diffset
                elif temp_diffset.pk == interdiffset_or_id:
                    interdiffset = temp_diffset
                else:
                    assert False

        filediff = get_object_or_404(FileDiff, pk=filediff_id, diffset=diffset)

        base_filediff = None
        interfilediff = None

        if interfilediff_id and base_filediff_id:
            raise UserVisibleError(_(
                'Cannot generate an interdiff when base FileDiff ID is '
                'specified.'
            ))
        elif interfilediff_id:
            interfilediff = get_object_or_404(FileDiff, pk=interfilediff_id,
                                              diffset=interdiffset)
        elif base_filediff_id:
            base_filediff = get_object_or_404(FileDiff, pk=base_filediff_id,
                                              diffset=diffset)

            ancestors = filediff.get_ancestors(minimal=False)

            if base_filediff not in ancestors:
                raise UserVisibleError(_(
                    'The requested FileDiff (ID %s) is not a valid base '
                    'FileDiff for FileDiff %s.'
                ) % (base_filediff_id, filediff_id))

        assert diffset is not None

        # Store this so we don't end up causing an SQL query later when looking
        # this up.
        filediff.diffset = diffset

        diff_file = self._get_requested_diff_file(
            diffset=diffset,
            filediff=filediff,
            interdiffset=interdiffset,
            interfilediff=interfilediff,
            base_filediff=base_filediff,
            diff_settings=DiffSettings.create(request=self.request))

        if not diff_file:
            raise UserVisibleError(
                _('Internal error. Unable to locate file record for '
                  'filediff %s')
                % filediff.pk)

        return {
            'diffset': diffset,
            'interdiffset': interdiffset,
            'filediff': filediff,
            'diff_file': diff_file,
        }

    @deprecate_non_keyword_only_args(RemovedInReviewBoard10_0Warning)
    def create_renderer(
        self,
        *,
        context: Mapping[str, Any],
        renderer_settings: Mapping[str, Any],
        diff_file: SerializedDiffFile,
        request: HttpRequest,
        **kwargs,
    ) -> DiffRenderer:
        """Create the renderer for the diff.

        This calculates all the state and data needed for rendering, and
        constructs a DiffRenderer with that data. That renderer is then
        returned, ready for rendering.

        If there's an error in looking up the necessary information, this
        may raise a UserVisibleError (best case), or some other form of
        Exception.

        Version Changed:
            8.0:
            * Deprecated non-keyword arguments.
            * Added the ``request`` parameter.

        Args:
            context (dict):
                The template rendering context.

            renderer_settings (dict):
                The diff renderer settings.settings

            diff_file (reviewboard.diffviewer.diffutils.SerializedDiffFile):
                The information on the diff file to render.

            request (django.http.HttpRequest):
                The request from the client.

                Version Added:
                    8.0

            **kwargs (dict):
                Keyword arguments, for future expansion.

        Returns:
            reviewboard.diffviewer.renderers.DiffRenderer:
            The resulting diff renderer.
        """
        return get_diff_renderer(
            diff_file,
            extra_context=context,
            template_name=self.template_name,
            **renderer_settings)

    def get_context_data(self, *args, **kwargs):
        """Returns context data used for rendering the view.

        This can be overridden by subclasses to provide additional data for the
        view.
        """
        return {}

    def _get_renderer_settings(self, chunk_index=None, **kwargs):
        """Calculate the render settings for the display of a diff.

        This will calculate settings based on user preferences and URL
        parameters. It does not calculate the state of any DiffSets or
        FileDiffs.
        """
        try:
            lines_of_context = self.request.GET.get('lines-of-context', '')
            lines_of_context = [int(i) for i in lines_of_context.split(',', 1)]
        except (TypeError, ValueError):
            lines_of_context = None

        if chunk_index is not None:
            try:
                chunk_index = int(chunk_index)
            except (TypeError, ValueError):
                chunk_index = None

        if lines_of_context:
            collapse_all = True
        elif chunk_index is not None:
            # If we're currently expanding part of a chunk, we want to render
            # the entire chunk without any lines collapsed. In the case of
            # showing a range of lines, we're going to get all chunks and then
            # only show the range. This is so that we won't have separate
            # cached entries for each range.
            collapse_all = False
        else:
            collapse_all = get_collapse_diff(self.request)

        show_deleted = (self.request.GET.get('show-deleted') == '1')

        return {
            'chunk_index': chunk_index,
            'collapse_all': collapse_all,
            'diff_settings': DiffSettings.create(request=self.request),
            'lines_of_context': lines_of_context,
            'show_deleted': show_deleted,
        }

    def _get_requested_diff_file(
        self,
        *,
        diffset: DiffSet,
        filediff: Optional[FileDiff],
        interdiffset: Optional[DiffSet],
        interfilediff: Optional[FileDiff],
        base_filediff: Optional[FileDiff],
        diff_settings: Optional[DiffSettings],
    ) -> Optional[SerializedDiffFile]:
        """Fetch information on the requested diff.

        This will look up information on the diff that's to be rendered
        and return it, if found. It may also augment it with additional
        data.

        The file will not contain chunk information. That must be specifically
        populated later.

        Version Changed:
            7.0.4:
            * Made arguments keyword-only.
            * Added the ``diff_settings`` argument.

        Args:
            diffset (reviewboard.diffviewer.models.diffset.DiffSet):
                The diffset containing the files to return.

            filediff (reviewboard.diffviewer.models.filediff.FileDiff,
                      optional):
                A specific file in the diff to return information for.

            interdiffset (reviewboard.diffviewer.models.diffset.DiffSet,
                          optional):
                A second diffset used for an interdiff range.

            interfilediff (reviewboard.diffviewer.models.filediff.FileDiff,
                           optional):
                A second specific file in ``interdiffset`` used to return
                information for. This should be provided if ``filediff`` and
                ``interdiffset`` are both provided. If it's ``None`` in this
                case, then the diff will be shown as reverted for this file.

                This may not be provided if ``base_filediff`` is provided.

            base_filediff (reviewbaord.diffviewer.models.filediff.FileDiff,
                           optional):
                The base FileDiff to use.

                This may only be provided if ``filediff`` is provided and
                ``interfilediff`` is not.

            diff_settings (reviewboard.diffviewer.settings.DiffSettings,
                           optional):
                The diff settings object. This will become mandatory in Review
                Board 9.0.

                Version Added:
                    7.0.4

        Returns:
            SerializedDiffFile:
            The serialized file information.
        """
        files = get_diff_files(diffset=diffset,
                               interdiffset=interdiffset,
                               filediff=filediff,
                               interfilediff=interfilediff,
                               base_filediff=base_filediff,
                               request=self.request,
                               diff_settings=diff_settings)

        if files:
            diff_file = files[0]

            try:
                diff_file['index'] = int(self.request.GET['index'])
            except (KeyError, ValueError):
                pass

            return diff_file

        return None


class DownloadPatchErrorBundleView(DiffFragmentView):
    """A view to download the patch error bundle.

    This view allows users to download a bundle containing data to help debug
    issues when a patch fails to apply. The bundle will contain the diff, the
    original file (as returned by the SCMTool), and the rejects file, if
    applicable.
    """

    def get(self, request, *args, **kwargs):
        """Handle GET requests for this view.

        This will create the renderer for the diff fragment and render it in
        order to get the PatchError information. It then returns a response
        with a zip file containing all the debug data.

        If no PatchError occurred, this will return a 404.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            *args (tuple):
                Additional positional arguments for the view.

            **kwargs (dict):
                Additional keyword arguments for the view.

        Returns:
            django.http.HttpResponse:
            A response containing the data bundle.
        """
        try:
            renderer_settings = self._get_renderer_settings(**kwargs)
            etag = self.make_etag(renderer_settings=renderer_settings,
                                  request=request, **kwargs)

            if etag_if_none_match(request, etag):
                return HttpResponseNotModified()

            diff_info = self.process_diffset_info(**kwargs)
        except Http404:
            return HttpResponseNotFound()
        except Exception as e:
            logger.exception(
                '%s.get: Error when processing diffset info for filediff '
                'ID=%s, interfilediff ID=%s, chunk_index=%s: %s',
                self.__class__.__name__,
                kwargs.get('filediff_id'),
                kwargs.get('interfilediff_id'),
                kwargs.get('chunk_index'),
                e,
                extra={'request': request})
            return HttpResponseServerError()

        kwargs.update(diff_info)

        try:
            context = self.get_context_data(**kwargs)

            renderer = self.create_renderer(
                request=request,
                context=context,
                renderer_settings=renderer_settings,
                **kwargs)
            renderer.render_to_response(request)
        except PatchError as e:
            patch_error = e
        except Exception as e:
            logger.exception(
                '%s.get: Error when rendering diffset for filediff ID=%s, '
                'interfilediff ID=%s, chunk_index=%s: %s',
                self.__class__.__name__,
                kwargs.get('filediff_id'),
                kwargs.get('interfilediff_id'),
                kwargs.get('chunk_index'),
                e,
                extra={'request': request})
            return HttpResponseServerError()
        else:
            return HttpResponseNotFound()

        zip_data = BytesIO()

        with ZipFile(zip_data, 'w') as zipfile:
            basename = os.path.basename(patch_error.filename)
            zipfile.writestr('%s.orig' % basename, patch_error.orig_file)
            zipfile.writestr('%s.diff' % basename, patch_error.diff)

            if patch_error.rejects:
                zipfile.writestr('%s.rej' % basename, patch_error.rejects)

            if patch_error.new_file:
                zipfile.writestr('%s.new' % basename, patch_error.new_file)

        rsp = HttpResponse(zip_data.getvalue(),
                           content_type='application/zip')
        rsp['Content-Disposition'] = \
            'attachment; filename=%s.zip' % basename

        return rsp


def exception_traceback_string(request, e, template_name, extra_context={}):
    context = {'error': e}
    context.update(extra_context)

    if not isinstance(e, UserVisibleError):
        context['trace'] = traceback.format_exc()

    return render_to_string(template_name=template_name,
                            context=context,
                            request=request)


def exception_traceback(request, e, template_name, extra_context={}):
    return HttpResponseServerError(
        exception_traceback_string(request, e, template_name, extra_context))



from __future__ import annotations

import logging
import uuid
from itertools import chain
from typing import (Any, Generic, Iterable, Mapping, Optional, Sequence,
                    TYPE_CHECKING, cast)

from django.db import models
from django.template.loader import get_template, render_to_string
from django.urls import NoReverseMatch
from django.utils.functional import cached_property
from django.utils.html import escape, format_html, format_html_join
from django.utils.safestring import mark_safe
from django.utils.translation import gettext, gettext_lazy as _
from typing_extensions import TypeVar

from reviewboard.accounts.models import User
from reviewboard.attachments.models import FileAttachment
from reviewboard.diffviewer.diffutils import get_sorted_filediffs
from reviewboard.diffviewer.models import DiffCommit, DiffSet
from reviewboard.reviews.fields import (BaseCommaEditableField,
                                        BaseEditableField,
                                        BaseReviewRequestField,
                                        BaseReviewRequestFieldSet,
                                        BaseTextAreaField,
                                        TFieldValue)
from reviewboard.reviews.models import (Group, ReviewRequest,
                                        ReviewRequestDraft,
                                        Screenshot)
from reviewboard.scmtools.models import Repository
from reviewboard.site.urlresolvers import local_site_reverse

if TYPE_CHECKING:
    from django.utils.safestring import SafeString
    from djblets.webapi.responses import WebAPIResponsePayload

    from reviewboard.changedescs.models import ChangeDescription
    from reviewboard.reviews.detail import ReviewRequestPageData
    from reviewboard.reviews.fields import ReviewRequestFieldChangeEntrySection
    from reviewboard.reviews.models.base_review_request_details import \
        BaseReviewRequestDetails

    FieldMixinParent = BaseReviewRequestField
else:
    FieldMixinParent = Generic


logger = logging.getLogger(__name__)


#: A generic type for a model passed to a review request field.
#:
#: Version Added:
#:     8.0
TModel = TypeVar('TModel', bound=models.Model)


class BuiltinFieldMixin(FieldMixinParent[TFieldValue]):
    """Mixin for built-in fields.

    This overrides some functions to work with native fields on a
    ReviewRequest or ReviewRequestDraft, rather than working with those
    stored in extra_data.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Initialize the field.

        Args:
            *args (tuple):
                Positional arguments to pass through to the superclass.

            **kwargs (dict):
                Keyword arguments to pass through to the superclass.
        """
        super().__init__(*args, **kwargs)

        field_id = self.field_id
        assert field_id is not None

        review_request_details = self.review_request_details

        if (not hasattr(review_request_details, field_id) and
            isinstance(review_request_details, ReviewRequestDraft)):
            # This field only exists in ReviewRequest, and not in
            # the draft, so we're going to work there instead.
            review_request_details = \
                review_request_details.get_review_request()

    def load_value(
        self,
        review_request_details: BaseReviewRequestDetails,
    ) -> Optional[TFieldValue]:
        """Load a value from the review request or draft.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            object:
            The loaded value.
        """
        field_id = self.field_id
        assert field_id

        value = getattr(review_request_details, field_id)

        if isinstance(value, models.Manager):
            value = list(value.all())

        return cast(Optional[TFieldValue], value)

    def save_value(
        self,
        value: Optional[TFieldValue],
    ) -> None:
        """Save the value in the review request or draft.

        Args:
            value (object):
                The new value for the field.
        """
        field_id = self.field_id
        assert field_id

        field = getattr(self.review_request_details, field_id)

        # ManyRelatedManager cannot be set with a simple assignment, so we need
        # to use .set() for that. Other field types can use setattr.
        if isinstance(field, models.Manager):
            field.set(value)
        else:
            setattr(self.review_request_details, field_id, value)


class BuiltinTextAreaFieldMixin(BuiltinFieldMixin[str]):
    """Mixin for built-in text area fields.

    This will ensure that the text is always rendered in Markdown,
    no matter whether the source text is plain or Markdown. It will
    still escape the text if it's not in Markdown format before
    rendering.
    """

    def get_data_attributes(self) -> dict[str, Any]:
        """Return any data attributes to include in the element.

        Returns:
            dict:
            The data attributes to include in the element.
        """
        attrs = super().get_data_attributes()

        # This is already available in the review request state fed to the
        # page, so we don't need it in the data attributes as well.
        attrs.pop('raw-value', None)

        return attrs


class ReviewRequestPageDataMixin(FieldMixinParent[TFieldValue]):
    """Mixin for internal fields needing access to the page data.

    These are used by fields that operate on state generated when creating the
    review request page. The view handling that page makes a lot of queries,
    and stores the results. This mixin allows access to those results,
    preventing additional queries.

    The data structure is not meant to be public API, and this mixin should not
    be used by any classes outside this file.

    By default, this will not render or handle any value loading or change
    entry recording. Subclasses must implement those manually.
    """

    #: Whether the field should be rendered.
    should_render: bool = False

    ######################
    # Instance variables #
    ######################

    #: The data already queried for the review request page.
    data: Optional[ReviewRequestPageData]

    def __init__(
        self,
        review_request_details: BaseReviewRequestDetails,
        data: Optional[ReviewRequestPageData] = None,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the mixin.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request (or the active draft thereof). In practice
                this will either be a
                :py:class:`reviewboard.reviews.models.ReviewRequest` or a
                :py:class:`reviewboard.reviews.models.ReviewRequestDraft`.

            data (reviewboard.reviews.detail.ReviewRequestPageData):
                The data already queried for the review request page.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.
        """
        super().__init__(review_request_details, *args, **kwargs)

        self.data = data

    def load_value(
        self,
        review_request_details: BaseReviewRequestDetails,
    ) -> Optional[TFieldValue]:
        """Load a value from the review request or draft.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            object:
            The loaded value.
        """
        return None

    def record_change_entry(
        self,
        changedesc: ChangeDescription,
        old_value: Any,
        new_value: Any,
    ) -> None:
        """Record information on the changed values in a ChangeDescription.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description to record the entry in.

            old_value (object):
                The old value of the field.

            new_value (object):
                The new value of the field.
        """
        pass


class BaseCaptionsField(ReviewRequestPageDataMixin[str],
                        BaseReviewRequestField[str]):
    """Base class for rendering captions for attachments.

    This serves as a base for FileAttachmentCaptionsField and
    ScreenshotCaptionsField. It provides the base rendering and
    for caption changes on file attachments or screenshots.
    """

    obj_map_attr: Optional[str] = None
    caption_object_field: Optional[str] = None

    change_entry_renders_inline = False

    def render_change_entry_html(
        self,
        info: object,
    ) -> SafeString:
        """Render a change entry to HTML.

        This function is expected to return safe, valid HTML. Any values
        coming from a field or any other form of user input must be
        properly escaped.

        Args:
            info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            django.utils.safestring.SafeString:
            The HTML representation of the change entry.
        """
        assert isinstance(info, dict)

        data = self.data
        obj_map_attr = self.obj_map_attr

        assert data is not None
        assert obj_map_attr is not None

        render_item = super().render_change_entry_html
        obj_map = getattr(data, obj_map_attr)

        s = ['<table class="caption-changed">']

        for id_str, caption in info.items():
            obj = obj_map[int(id_str)]

            s.append(format_html(
                '<tr>'
                ' <th><a href="{url}">{filename}</a>:</th>'
                ' <td>{caption}</td>'
                '</tr>',
                url=obj.get_absolute_url(),
                filename=obj.filename,
                caption=mark_safe(render_item(caption))))

        s.append('</table>')

        return mark_safe(''.join(s))

    def serialize_change_entry(
        self,
        changedesc: ChangeDescription,
    ) -> Sequence[WebAPIResponsePayload]:
        """Serialize a change entry for public consumption.

        This will output a version of the change entry for use in the API.
        It can be the same content stored in the
        :py:class:`~reviewboard.changedescs.models.ChangeDescription`, but
        does not need to be.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description whose field is to be serialized.

        Returns:
            list:
            An appropriate serialization for the field.
        """
        data = changedesc.fields_changed[self.field_id]

        model = self.model
        assert model is not None

        return [
            {
                'old': data[str(obj.pk)]['old'][0],
                'new': data[str(obj.pk)]['new'][0],
                self.caption_object_field: obj,
            }
            for obj in model.objects.filter(pk__in=data.keys())
        ]


class BaseModelListEditableField(BaseCommaEditableField[TModel]):
    """Base class for editable comma-separated list of model instances.

    This is used for built-in classes that work with ManyToManyFields.
    """

    model_name_attr: Optional[str] = None

    def has_value_changed(
        self,
        old_value: Sequence[TModel],
        new_value: Sequence[TModel],
    ) -> bool:
        """Return whether the value has changed.

        Args:
            old_value (object):
                The old value of the field.

            new_value (object):
                The new value of the field.

        Returns:
            bool:
            Whether the value of the field has changed.
        """
        old_values = {obj.pk for obj in old_value}
        new_values = {obj.pk for obj in new_value}

        return old_values != new_values

    def record_change_entry(
        self,
        changedesc: ChangeDescription,
        old_value: Sequence[TModel],
        new_value: Sequence[TModel],
    ) -> None:
        """Record information on the changed values in a ChangeDescription.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description to record the entry in.

            old_value (object):
                The old value of the field.

            new_value (object):
                The new value of the field.
        """
        assert self.field_id

        changedesc.record_field_change(field=self.field_id,
                                       old_value=old_value,
                                       new_value=new_value,
                                       name_field=self.model_name_attr)

    def render_change_entry_item_html(
        self,
        info: object,
        item: tuple[str, str, int],
    ) -> SafeString:
        """Render an item for change description HTML.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            item (object):
                The value of the item.

        Returns:
            django.utils.safestring.SafeString:
            The rendered change entry.
        """
        label, url, pk = item

        if url:
            return format_html('<a href="{}">{}</a>', url, label)
        else:
            return escape(label)

    def save_value(
        self,
        value: Optional[Sequence[TModel]],
    ) -> None:
        """Save the value in the review request or draft.

        Args:
            value (object):
                The new value for the field.
        """
        assert self.field_id

        setattr(self.review_request_details, self.field_id, value)


class StatusField(BuiltinFieldMixin[str],
                  BaseReviewRequestField[str]):
    """The Status field on a review request."""

    field_id = 'status'
    label = _('Status')
    is_required = True

    #: Whether the field should be rendered.
    should_render = False

    def get_change_entry_sections_html(
        self,
        info: object,
    ) -> Sequence[ReviewRequestFieldChangeEntrySection]:
        """Return sections of change entries with titles and rendered HTML.

        Because the status field is specially handled, this just returns an
        empty list.

        Args:
            info (dict, unused):
                A dictionary describing how the field has changed.

        Returns:
            list:
            An empty list.
        """
        return []


class SummaryField(BuiltinFieldMixin, BaseEditableField):
    """The Summary field on a review request."""

    field_id = 'summary'
    label = _('Summary')
    is_required = True
    tag_name = 'h1'

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.SummaryFieldView'


class DescriptionField(BuiltinTextAreaFieldMixin, BaseTextAreaField):
    """The Description field on a review request."""

    field_id = 'description'
    label = _('Description')
    is_required = True

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.DescriptionFieldView'

    def is_text_markdown(
        self,
        value: Optional[str],
    ) -> bool:
        """Return whether the description uses Markdown.

        Args:
            value (str, unused):
                The field's text.

        Returns:
            bool:
            True if the description field should be formatted using Markdown.
        """
        return self.review_request_details.description_rich_text


class TestingDoneField(BuiltinTextAreaFieldMixin, BaseTextAreaField):
    """The Testing Done field on a review request."""

    field_id = 'testing_done'
    label = _('Testing Done')

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.TestingDoneFieldView'

    def is_text_markdown(
        self,
        value: Optional[str],
    ) -> bool:
        """Return whether the description uses Markdown.

        Args:
            value (str, unused):
                The field's text.

        Returns:
            bool:
            True if the description field should be formatted using Markdown.
        """
        return self.review_request_details.testing_done_rich_text


class OwnerField(BuiltinFieldMixin[User],
                 BaseEditableField[User]):
    """The Owner field on a review request."""

    field_id = 'submitter'
    label = _('Owner')
    model = User
    model_name_attr = 'username'
    is_required = True

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.OwnerFieldView'

    def render_value(
        self,
        value: Optional[User],
    ) -> SafeString:
        """Render the value in the field.

        Args:
            value (django.contrib.auth.models.User):
                The value to render.

        Returns:
            django.utils.safestring.SafeString:
            The rendered value.
        """
        assert value is not None

        user = value
        request = self.request
        review_request_details = self.review_request_details

        assert request is not None
        assert review_request_details is not None

        return format_html(
            '<a class="user" href="{0}">{1}</a>',
            local_site_reverse(
                'user',
                local_site=review_request_details.local_site,
                args=[user]),
            user.get_profile().get_display_name(request.user))

    def record_change_entry(
        self,
        changedesc: ChangeDescription,
        old_value: User,
        new_value: User,
    ) -> None:
        """Record information on the changed values in a ChangeDescription.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description to record the entry in.

            old_value (django.contrib.auth.models.User):
                The old value of the field.

            new_value (django.contrib.auth.models.User):
                The new value of the field.
        """
        field_id = self.field_id
        assert field_id

        changedesc.record_field_change(
            field=field_id,
            old_value=old_value,
            new_value=new_value,
            name_field=self.model_name_attr,
            build_url_func=lambda user: local_site_reverse(
                'user',
                local_site=self.review_request_details.local_site,
                kwargs={
                    'username': user.username,
                }))

    def render_change_entry_value_html(
        self,
        info: object,
        value: tuple[str, str, int],
    ) -> SafeString:
        """Render the value for a change description string to HTML.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            item (object):
                The value of the field.

        Returns:
            django.utils.safestring.SafeString:
            The rendered change entry.
        """
        label, url, pk = value

        if url:
            return format_html('<a href="{}">{}</a>', url, label)
        else:
            return escape(label)

    def serialize_change_entry(
        self,
        changedesc: ChangeDescription,
    ) -> WebAPIResponsePayload:
        """Serialize a change entry for public consumption.

        This will output a version of the change entry for use in the API.
        It can be the same content stored in the
        :py:class:`~reviewboard.changedescs.models.ChangeDescription`, but
        does not need to be.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description whose field is to be serialized.

        Returns:
            dict:
            An appropriate serialization for the field.
        """
        entry = super().serialize_change_entry(changedesc)
        assert isinstance(entry, dict)

        return {
            key: value[0]
            for key, value in entry.items()
        }


class RepositoryField(BuiltinFieldMixin[Repository],
                      BaseReviewRequestField[Repository]):
    """The Repository field on a review request."""

    field_id = 'repository'
    label = _('Repository')
    model = Repository

    @property
    def should_render(self) -> bool:
        """Whether the field should be rendered."""
        review_request = self.review_request_details.get_review_request()

        return review_request.repository_id is not None


class BranchField(BuiltinFieldMixin[str],
                  BaseEditableField[str]):
    """The Branch field on a review request."""

    field_id = 'branch'
    label = _('Branch')

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.BranchFieldView'


class BugsField(BuiltinFieldMixin[Sequence[str]],
                BaseCommaEditableField[str]):
    """The Bugs field on a review request."""

    field_id = 'bugs_closed'
    label = _('Bugs')

    one_line_per_change_entry = False

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.BugsFieldView'

    def load_value(
        self,
        review_request_details: BaseReviewRequestDetails,
    ) -> Sequence[str]:
        """Load a value from the review request or draft.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            object:
            The loaded value.
        """
        return review_request_details.get_bug_list()

    def save_value(
        self,
        value: Optional[Sequence[str]],
    ) -> None:
        """Save the value in the review request or draft.

        Args:
            value (object):
                The new value for the field.
        """
        assert self.field_id

        setattr(self.review_request_details, self.field_id, (
            ', '.join(value or [])
            .replace('\n', '')
            .replace('\r', '')
        ))

    def render_item(
        self,
        item: str,
    ) -> SafeString:
        """Render an item from the list.

        Args:
            item (object):
                The item to render.

        Returns:
            django.utils.safestring.SafeString:
            The rendered item.
        """
        bug_url = self._get_bug_url(item)

        if bug_url:
            return format_html('<a class="bug" href="{url}">{id}</a>',
                               url=bug_url,
                               id=item)
        else:
            return escape(item)

    def render_change_entry_item_html(
        self,
        info: object,
        item: tuple[str, str, int],
    ) -> SafeString:
        """Render an item for change description HTML.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            item (object):
                The value of the item.

        Returns:
            django.utils.safestring.SafeString:
            The rendered change entry.
        """
        return self.render_item(item[0])

    def _get_bug_url(
        self,
        bug_id: str,
    ) -> Optional[str]:
        """Return the URL to link to a specific bug.

        Args:
            bug_id (str):
                The ID of the bug to link to.

        Returns:
            str:
            The link to view the bug in the bug tracker, if available.
        """
        review_request_details = self.review_request_details
        review_request = review_request_details.get_review_request()
        repository = review_request_details.repository
        local_site_name: Optional[str] = None
        bug_url: Optional[str] = None

        if review_request.local_site:
            local_site_name = review_request.local_site.name

        try:
            if (repository and
                repository.bug_tracker and
                '%s' in repository.bug_tracker):
                bug_url = local_site_reverse(
                    'bug_url', local_site_name=local_site_name,
                    args=[review_request.display_id, bug_id])
        except NoReverseMatch:
            pass

        return bug_url


class DependsOnField(BuiltinFieldMixin[Sequence[ReviewRequest]],
                     BaseModelListEditableField[ReviewRequest]):
    """The Depends On field on a review request."""

    field_id = 'depends_on'
    label = _('Depends On')
    model = ReviewRequest
    model_name_attr = 'summary'

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.DependsOnFieldView'

    def render_change_entry_item_html(
        self,
        info: object,
        item: tuple[str, str, int],
    ) -> SafeString:
        """Render an item for change description HTML.

        Args:
            info (dict):
                A dictionary describing how the field has changed.

            item (object):
                The value of the item.

        Returns:
            django.utils.safestring.SafeString:
            The rendered change entry.
        """
        obj = ReviewRequest.objects.get(pk=item[2])

        rendered_item = format_html(
            '<a href="{url}">{id} - {summary}</a>',
            url=obj.get_absolute_url(),
            id=obj.pk,
            summary=obj.summary)

        if obj.status in (ReviewRequest.SUBMITTED,
                          ReviewRequest.DISCARDED):
            rendered_item = format_html('<s>{}</s>', rendered_item)

        return rendered_item

    def render_item(
        self,
        item: ReviewRequest,
    ) -> SafeString:
        """Render an item from the list.

        Args:
            item (object):
                The item to render.

        Returns:
            django.utils.safestring.SafeString:
            The rendered item.
        """
        rendered_item = format_html(
            '<a href="{url}" title="{summary}"'
            '   class="review-request-link">{id}</a>',
            url=item.get_absolute_url(),
            summary=item.summary,
            id=item.display_id)

        if item.status in (ReviewRequest.SUBMITTED,
                           ReviewRequest.DISCARDED):
            rendered_item = format_html('<s>{}</s>', rendered_item)

        return rendered_item


class BlocksField(BuiltinFieldMixin[Sequence[ReviewRequest]],
                  BaseReviewRequestField[Sequence[ReviewRequest]]):
    """The Blocks field on a review request."""

    field_id = 'blocks'
    label = _('Blocks')
    model = ReviewRequest

    def load_value(
        self,
        review_request_details: BaseReviewRequestDetails,
    ) -> Optional[Sequence[ReviewRequest]]:
        """Load a value from the review request or draft.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            object:
            The loaded value.
        """
        return review_request_details.get_review_request().get_blocks()

    @property
    def should_render(self) -> bool:
        """Whether the field should be rendered."""
        return bool(self.value)

    def render_value(
        self,
        value: Optional[Sequence[ReviewRequest]],
    ) -> SafeString:
        """Render the value in the field.

        Args:
            blocks (list):
                The value to render.

        Returns:
            django.utils.safestring.SafeString:
            The rendered value.
        """
        return format_html_join(
            ', ',
            '<a href="{0}" class="review-request-link">{1}</a>',
            [
                (item.get_absolute_url(), item.display_id)
                for item in (value or [])
            ])


class ChangeField(BuiltinFieldMixin[str],
                  BaseReviewRequestField[str]):
    """The Change field on a review request.

    This is shown for repositories supporting changesets. The change
    number is similar to a commit ID, with the exception that it's only
    ever stored on the ReviewRequest and never changes.

    If both ``changenum`` and ``commit_id`` are provided on the review
    request, only this field will be shown, as both are likely to have
    values.
    """

    field_id = 'changenum'
    label = _('Change')

    def load_value(
        self,
        review_request_details: BaseReviewRequestDetails,
    ) -> str:
        """Load a value from the review request or draft.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            str:
            The loaded value.
        """
        return review_request_details.get_review_request().changenum

    @property
    def should_render(self) -> bool:
        """Whether the field should be rendered."""
        return bool(self.value)

    def render_value(
        self,
        value: Optional[str],
    ) -> SafeString:
        """Render the value in the field.

        Args:
            changenum (str):
                The value to render.

        Returns:
            django.utils.safestring.SafeString:
            The rendered value.
        """
        review_request = self.review_request_details.get_review_request()

        is_pending, changenum = review_request.changeset_is_pending(value)

        if is_pending:
            fmt = gettext('{} (pending)')
        else:
            fmt = '{}'

        return format_html(fmt, changenum)


class CommitField(BuiltinFieldMixin[str],
                  BaseReviewRequestField[str]):
    """The Commit field on a review request.

    This displays the ID of the commit the review request is representing.

    Since the ``commit_id`` and ``changenum`` fields are both populated, we
    let ChangeField take precedence. It knows how to render information based
    on a changeset ID.
    """

    field_id = 'commit_id'
    label = _('Commit')
    can_record_change_entry = True
    tag_name = 'span'

    @property
    def should_render(self) -> bool:
        """Whether the field should be rendered."""
        return (bool(self.value) and
                not self.review_request_details.get_review_request().changenum)

    def render_value(
        self,
        value: Optional[str],
    ) -> SafeString:
        """Render the value in the field.

        Args:
            commit_id (str):
                The value to render.

        Returns:
            django.utils.safestring.SafeString:
            The rendered value.
        """
        assert value is not None

        # Abbreviate SHA-1s
        commit_id = value

        if len(commit_id) == 40:
            abbrev_commit_id = commit_id[:7] + '...'

            return format_html('<span title="{}">{}</span>',
                               commit_id,
                               abbrev_commit_id)
        else:
            return escape(commit_id)


class DiffField(ReviewRequestPageDataMixin[DiffSet],
                BuiltinFieldMixin[DiffSet],
                BaseReviewRequestField[DiffSet]):
    """Represents a newly uploaded diff on a review request.

    This is not shown as an actual displayable field on the review request
    itself. Instead, it is used only during the ChangeDescription population
    and processing steps.
    """

    field_id = 'diff'
    label = _('Diff')

    can_record_change_entry = True

    MAX_FILES_PREVIEW = 8

    def render_change_entry_html(
        self,
        info: object,
    ) -> SafeString:
        """Render a change entry to HTML.

        This function is expected to return safe, valid HTML. Any values
        coming from a field or any other form of user input must be
        properly escaped.

        Args:
            info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            django.utils.safestring.SafeString:
            The HTML representation of the change entry.
        """
        assert isinstance(info, dict)

        added_diff_info = info['added'][0]
        review_request = self.review_request_details.get_review_request()

        data = self.data
        assert data is not None

        try:
            diffset = data.diffsets_by_id[added_diff_info[2]]
        except KeyError:
            # If a published revision of a diff has been deleted from the
            # database, this will explode. Just return a blank string for this,
            # so that it doesn't show a traceback.
            return mark_safe('')

        diff_revision = diffset.revision
        past_revision = diff_revision - 1
        diff_url = added_diff_info[1]

        s: list[SafeString] = []

        # Fetch the total number of inserts/deletes. These will be shown
        # alongside the diff revision.
        counts = diffset.get_total_line_counts()
        raw_insert_count = counts.get('raw_insert_count', 0)
        raw_delete_count = counts.get('raw_delete_count', 0)

        line_counts = []

        if raw_insert_count > 0:
            line_counts.append('<span class="insert-count">+%d</span>'
                               % raw_insert_count)

        if raw_delete_count > 0:
            line_counts.append('<span class="delete-count">-%d</span>'
                               % raw_delete_count)

        # Display the label, URL, and line counts for the diff.
        if line_counts:
            s.append(format_html(
                '<p class="diff-changes">'
                ' <a href="{url}">{label}</a>'
                ' <span class="line-counts">({line_counts})</span>'
                '</p>',
                url=diff_url,
                label=_('Revision %s') % diff_revision,
                line_counts=mark_safe(' '.join(line_counts))))
        else:
            s.append(format_html(
                '<p class="diff-changes">'
                ' <a href="{url}">{label}</a>'
                '</p>',
                url=diff_url,
                label=_('Revision %s') % diff_revision))

        if past_revision > 0:
            # This is not the first diff revision. Include an interdiff link.
            interdiff_url = local_site_reverse(
                'view-interdiff',
                local_site=review_request.local_site,
                args=[
                    review_request.display_id,
                    past_revision,
                    diff_revision,
                ])

            s.append(format_html(
                '<p><a href="{url}">{text}</a>',
                url=interdiff_url,
                text=_('Show changes')))

        file_count = len(diffset.cumulative_files)

        if file_count > 0:
            # Begin displaying the list of files modified in this diff.
            # It will be capped at a fixed number (MAX_FILES_PREVIEW).
            s += [
                mark_safe('<div class="diff-index">'),
                mark_safe(' <table>'),
            ]

            # We want a sorted list of filediffs, but tagged with the order in
            # which they come from the database, so that we can properly link
            # to the respective files in the diff viewer.
            files = get_sorted_filediffs(enumerate(diffset.cumulative_files),
                                         key=lambda i: i[1])

            for i, filediff in files[:self.MAX_FILES_PREVIEW]:
                counts = filediff.get_line_counts()

                data_attrs = [
                    'data-%s="%s"' % (attr.replace('_', '-'), counts[attr])
                    for attr in ('insert_count', 'delete_count',
                                 'replace_count', 'total_line_count')
                    if counts.get(attr) is not None
                ]

                s.append(format_html(
                    '<tr {data_attrs}>'
                    ' <td class="diff-file-icon"></td>'
                    ' <td class="diff-file-info">'
                    '  <a href="{url}">{filename}</a>'
                    ' </td>'
                    '</tr>',
                    data_attrs=mark_safe(' '.join(data_attrs)),
                    url=diff_url + '#%d' % i,
                    filename=filediff.source_file))

            num_remaining = file_count - self.MAX_FILES_PREVIEW

            if num_remaining > 0:
                # There are more files remaining than we've shown, so show
                # the count.
                s.append(format_html(
                    '<tr>'
                    ' <td></td>'
                    ' <td class="diff-file-info">{text}</td>'
                    '</tr>',
                    text=_('%s more') % num_remaining))

            s += [
                mark_safe(' </table>'),
                mark_safe('</div>'),
            ]

        return mark_safe(''.join(s))

    def has_value_changed(
        self,
        old_value: Optional[DiffSet],
        new_value: Optional[DiffSet],
    ) -> bool:
        """Return whether the value has changed.

        Args:
            old_value (object):
                The old value of the field.

            new_value (object):
                The new value of the field.

        Returns:
            bool:
            Whether the value of the field has changed.
        """
        # If there's a new diffset at all (in new_value), then it passes
        # the test.
        return new_value is not None

    def load_value(
        self,
        review_request_details: BaseReviewRequestDetails,
    ) -> Optional[DiffSet]:
        """Load a value from the review request or draft.

        Args:
            review_request_details (reviewboard.reviews.models.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            object:
            The loaded value.
        """
        # This will be None for a ReviewRequest, and may have a value for
        # ReviewRequestDraft if a new diff was attached.
        return getattr(review_request_details, 'diffset', None)

    def save_value(
        self,
        value: Optional[DiffSet],
    ) -> None:
        """Save the value in the review request or draft.

        Args:
            value (object):
                The new value for the field.
        """
        # The diff is a fake field that doesn't actually exist on the review
        # request, so it deosn't make sense to save.
        pass

    def record_change_entry(
        self,
        changedesc: ChangeDescription,
        old_value: DiffSet,
        new_value: DiffSet,
    ) -> None:
        """Record information on the changed values in a ChangeDescription.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description to record the entry in.

            old_value (object):
                The old value of the field.

            new_value (object):
                The new value of the field.
        """
        diffset = new_value

        review_request = self.review_request_details.get_review_request()

        url = local_site_reverse(
            'view-diff-revision',
            local_site=review_request.local_site,
            args=[review_request.display_id, diffset.revision])

        changedesc.fields_changed['diff'] = {
            'added': [(
                _('Diff r%s') % diffset.revision,
                url,
                diffset.pk,
            )]
        }

    def serialize_change_entry(
        self,
        changedesc,
    ) -> WebAPIResponsePayload:
        """Serialize a change entry for public consumption.

        This will output a version of the change entry for use in the API.
        It can be the same content stored in the
        :py:class:`~reviewboard.changedescs.models.ChangeDescription`, but
        does not need to be.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description whose field is to be serialized.

        Returns:
            dict:
            An appropriate serialization for the field.
        """
        diffset_id = changedesc.fields_changed['diff']['added'][0][2]

        return {
            'added': DiffSet.objects.get(pk=diffset_id),
        }


class FileAttachmentCaptionsField(BaseCaptionsField):
    """Renders caption changes for file attachments.

    This is not shown as an actual displayable field on the review request
    itself. Instead, it is used only during the ChangeDescription rendering
    stage. It is not, however, used for populating entries in
    ChangeDescription.
    """

    field_id = 'file_captions'
    label = _('File Captions')
    obj_map_attr = 'file_attachments_by_id'
    model = FileAttachment
    caption_object_field = 'file_attachment'


class FileAttachmentsField(
    ReviewRequestPageDataMixin[Sequence[FileAttachment]],
    BuiltinFieldMixin[Sequence[FileAttachment]],
    BaseCommaEditableField[FileAttachment],
):
    """Renders removed or added file attachments.

    This is not shown as an actual displayable field on the review request
    itself. Instead, it is used only during the ChangeDescription rendering
    stage. It is not, however, used for populating entries in
    ChangeDescription.
    """

    field_id = 'files'
    label = _('Files')
    model = FileAttachment

    thumbnail_template = 'reviews/changedesc_file_attachment.html'

    def get_change_entry_sections_html(
        self,
        info: Mapping[str, Any],
    ) -> Sequence[ReviewRequestFieldChangeEntrySection]:
        """Return sections of change entries with titles and rendered HTML.

        Args:
            info (dict):
                A dictionary describing how the field has changed. This is
                guaranteed to have ``new`` and ``old`` keys, but may also
                contain ``added`` and ``removed`` keys as well.

        Returns:
            list of dict:
            A list of the change entry sections.
        """
        assert isinstance(info, dict)

        sections: list[ReviewRequestFieldChangeEntrySection] = []

        if 'removed' in info:
            sections.append({
                'title': _('Removed Files'),
                'rendered_html': mark_safe(
                    self.render_change_entry_html(info['removed'])),
            })

        if 'added' in info:
            sections.append({
                'title': _('Added Files'),
                'rendered_html': mark_safe(
                    self.render_change_entry_html(info['added'])),
            })

        return sections

    def render_change_entry_html(
        self,
        info: Sequence[tuple[str, str, int]],
    ) -> SafeString:
        """Render a change entry to HTML.

        This function is expected to return safe, valid HTML. Any values
        coming from a field or any other form of user input must be
        properly escaped.

        Args:
            info (list):
                A list of the changed file attachments. Each item is a 3-tuple
                containing the ``caption``, ``filename``, and the ``pk`` of the
                file attachment in the database.

        Returns:
            django.utils.safestring.SafeString:
            The HTML representation of the change entry.
        """
        # Fetch the template ourselves only once and render it for each item,
        # instead of calling render_to_string() in the loop, so we don't
        # have to locate and parse/fetch from cache for every item.

        template = get_template(self.thumbnail_template)

        data = self.data
        assert data is not None

        items: list[SafeString] = []

        for caption, filename, pk in info:
            try:
                attachment = data.file_attachments_by_id[pk]
            except KeyError:
                try:
                    attachment = FileAttachment.objects.get(pk=pk)
                except FileAttachment.DoesNotExist:
                    continue

            items.append(template.render({
                'model_attrs': self.get_attachment_js_model_attrs(attachment),
                'uuid': uuid.uuid4(),
            }))

        if not items:
            return mark_safe('')

        return format_html(
            '<div class="rb-c-file-attachments">{}</div>',
            mark_safe(''.join(items)))

    def get_attachment_js_model_attrs(
        self,
        attachment: FileAttachment,
        draft: bool = False,
    ) -> dict[str, Any]:
        """Return attributes for the RB.FileAttachment JavaScript model.

        This will determine the right attributes to pass to an instance
        of :js:class:`RB.FileAttachment`, based on the provided file
        attachment.

        Args:
            attachment (reviewboard.attachments.models.FileAttachment):
                The file attachment to return attributes for.

            draft (bool, optional):
                Whether to return attributes for a draft version of the
                file attachment.

        Returns:
            dict:
            The resulting model attributes.
        """
        review_request = self.review_request_details.get_review_request()
        request = self.request

        model_attrs = {
            'canAccessReviewUI': attachment.is_review_ui_accessible_by(
                request.user),
            'downloadURL': attachment.get_absolute_url(),
            'filename': attachment.filename,
            'id': attachment.pk,
            'loaded': True,
            'publishedCaption': attachment.caption,
            'reviewURL': local_site_reverse(
                'file-attachment',
                kwargs={
                    'file_attachment_id': attachment.pk,
                    'review_request_id': review_request.display_id,
                },
                request=request),
            'revision': attachment.attachment_revision,
            'state': self.review_request_details.get_file_attachment_state(
                attachment).value,
            'thumbnailHTML': attachment.thumbnail,
        }

        if draft:
            caption = attachment.draft_caption
        else:
            caption = attachment.caption

        model_attrs['caption'] = caption

        if attachment.attachment_history_id:
            model_attrs['attachmentHistoryID'] = \
                attachment.attachment_history_id

        return model_attrs


class ScreenshotCaptionsField(BaseCaptionsField):
    """Renders caption changes for screenshots.

    This is not shown as an actual displayable field on the review request
    itself. Instead, it is used only during the ChangeDescription rendering
    stage. It is not, however, used for populating entries in
    ChangeDescription.
    """

    field_id = 'screenshot_captions'
    label = _('Screenshot Captions')
    obj_map_attr = 'screenshots_by_id'
    model = Screenshot
    caption_object_field = 'screenshot'


class ScreenshotsField(BaseCommaEditableField[Screenshot]):
    """Renders removed or added screenshots.

    This is not shown as an actual displayable field on the review request
    itself. Instead, it is used only during the ChangeDescription rendering
    stage. It is not, however, used for populating entries in
    ChangeDescription.
    """

    field_id = 'screenshots'
    label = _('Screenshots')
    model = Screenshot


class TargetGroupsField(BuiltinFieldMixin[Sequence[Group]],
                        BaseModelListEditableField[Group]):
    """The Target Groups field on a review request."""

    field_id = 'target_groups'
    label = _('Groups')
    model = Group
    model_name_attr = 'name'

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.TargetGroupsFieldView'

    def render_item(
        self,
        item: Group,
    ) -> SafeString:
        """Render an item from the list.

        Args:
            item (reviewboard.reviews.models.group.Group):
                The item to render.

        Returns:
            django.utils.safestring.SafeString:
            The rendered item.
        """
        return format_html('<a href="{}">{}</a>',
                           item.get_absolute_url(),
                           item.name)


class TargetPeopleField(BuiltinFieldMixin[Sequence[User]],
                        BaseModelListEditableField[User]):
    """The Target People field on a review request."""

    field_id = 'target_people'
    label = _('People')
    model = User
    model_name_attr = 'username'

    #: The class name for the JavaScript view representing this field.
    js_view_class = 'RB.ReviewRequestFields.TargetPeopleFieldView'

    def render_item(
        self,
        item: User,
    ) -> SafeString:
        """Render an item from the list.

        Args:
            item (django.contrib.auth.models.User):
                The item to render.

        Returns:
            django.utils.safestring.SafeString:
            The rendered item.
        """
        user = item
        extra_classes: list[str] = ['user']

        if not user.is_active:
            extra_classes.append('inactive')

        return format_html(
            '<a href="{0}" class="{1}">{2}</a>',
            local_site_reverse(
                'user',
                local_site=self.review_request_details.local_site,
                args=[user]),
            ' '.join(extra_classes),
            user.username)


class CommitListField(ReviewRequestPageDataMixin[DiffSet],
                      BaseReviewRequestField[DiffSet]):
    """The list of commits for a review request."""

    field_id = 'commit_list'
    label = _('Commits')

    is_editable = False

    js_view_class = 'RB.ReviewRequestFields.CommitListFieldView'

    @cached_property
    def review_request_created_with_history(self) -> bool:
        """Whether the associated review request was created with history."""
        return (
            self.review_request_details
            .get_review_request()
            .created_with_history
        )

    @property
    def should_render(self) -> bool:
        """Whether or not the field should be rendered.

        This field will only be rendered when the review request was created
        with history support. It is also hidden on the diff viewer page,
        because it substantially overlaps with the commit selector.
        """
        from reviewboard.urls import diffviewer_url_names
        url_name = self.request.resolver_match.url_name

        return (self.value is not None and
                self.review_request_created_with_history and
                url_name not in diffviewer_url_names)

    @property
    def can_record_change_entry(self) -> bool:
        """Whether or not the field can record a change entry.

        The field can only record a change entry when the review request has
        been created with history.
        """
        return self.review_request_created_with_history

    def load_value(
        self,
        review_request_details: BaseReviewRequestDetails,
    ) -> DiffSet:
        """Load a value from the review request or draft.

        Args:
            review_request_details (review_request_details.
                                    base_review_request_details.
                                    BaseReviewRequestDetails):
                The review request or draft.

        Returns:
            reviewboard.diffviewer.models.diffset.DiffSet:
            The DiffSet associated with the review request or draft.
        """
        return review_request_details.get_latest_diffset()

    def save_value(
        self,
        value: Optional[DiffSet],
    ) -> None:
        """Save a value to the review request.

        This is intentionally a no-op.

        Args:
            value (reviewboard.diffviewer.models.diffset.DiffSet, unused):
                The current DiffSet
        """
        pass

    def render_value(
        self,
        value: Optional[DiffSet],
    ) -> SafeString:
        """Render the field for the given value.

        Args:
            value (reviewboard.diffviewer.models.diffset.DiffSet):
                The loaded diffset.

        returns:
            django.utils.safestring.SafeString:
            The rendered value.
        """
        if not value:
            return mark_safe('')

        commits = list(value.commits.order_by('pk'))
        context = self._get_common_context(commits)
        context['commits'] = commits

        return render_to_string(
            template_name='reviews/commit_list_field.html',
            request=self.request,
            context=context)

    def has_value_changed(
        self,
        old_value: Optional[DiffSet],
        new_value: Optional[DiffSet],
    ) -> bool:
        """Return whether or not the value has changed.

        Args:
            old_value (reviewboard.diffviewer.models.diffset.DiffSet):
                The :py:class:`~reviewboard.diffviewer.
                models.diffset.DiffSet` from the review_request.

            new_value (reviewboard.diffviewer.models.diffset.DiffSet):
                The :py:class:`~reviewboard.diffviewer.
                models.diffset.DiffSet` from the draft.

        Returns:
            bool:
            Whether or not the value has changed.
        """
        return new_value is not None

    def record_change_entry(
        self,
        changedesc: ChangeDescription,
        old_value: Optional[DiffSet],
        new_value: DiffSet,
    ) -> None:
        """Record the old and new values for this field into the changedesc.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description to record the change into.

            old_value (reviewboard.diffviewer.models.diffset.DiffSet):
                The previous :py:class:`~reviewboard.diffviewer.models.
                diffset.DiffSet` from the review request.

            new_value (reviewboard.diffviewer.models.diffset.DiffSet):
                The new :py:class:`~reviewboard.diffviewer.models.diffset.
                DiffSet` from the draft.
        """
        changedesc.fields_changed[self.field_id] = {
            'old': old_value and old_value.pk,
            'new': new_value.pk,
        }

    def render_change_entry_html(
        self,
        info: Mapping[str, Any],
    ) -> SafeString:
        """Render the change entry HTML for this field.

        Args:
            info (dict):
                The change entry info for this field. See
                :py:meth:`record_change_entry` for the format.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        data = self.data
        assert data is not None

        commits = data.commits_by_diffset_id

        if info['old']:
            old_commits = commits[info['old']]
        else:
            old_commits = []

        new_commits = commits[info['new']]

        context = self._get_common_context(chain(old_commits, new_commits))
        context.update({
            'old_commits': old_commits,
            'new_commits': new_commits,
        })

        return render_to_string(
            template_name='reviews/changedesc_commit_list.html',
            request=self.request,
            context=context)

    def serialize_change_entry(
        self,
        changedesc: ChangeDescription,
    ) -> WebAPIResponsePayload:
        """Serialize the changed field entry for the web API.

        Args:
            changdesc (reviewboard.changedescs.models.ChangeDescription):
                The change description being serialized.

        Returns:
            dict:
            A JSON-serializable dictionary representing the change entry for
            this field.
        """
        info = changedesc.fields_changed[self.field_id]

        commits_by_diffset_id = DiffCommit.objects.by_diffset_ids(
            (info['old'], info['new']))

        return {
            key: [
                {
                    'author': commit.author_name,
                    'summary': commit.summary,
                }
                for commit in commits_by_diffset_id[info[key]]
            ]
            for key in ('old', 'new')
        }

    def _get_common_context(
        self,
        commits: Iterable[DiffCommit],
    ) -> dict[str, Any]:
        """Return common context for rending both change entries and the field.

        Args:
            commits (iterable of reviewboard.diffviewer.models.diffcommit.
                     DiffCommit):
                The commits to generate context for.

        Returns:
            dict:
            A dictionary of context.
        """
        review_request_details = self.review_request_details

        submitter_name = review_request_details.submitter.get_full_name()
        include_author_name = not submitter_name
        to_expand: set[int] = set()

        for commit in commits:
            if commit.author_name != submitter_name:
                include_author_name = True

            if commit.summary.strip() != commit.commit_message.strip():
                to_expand.add(commit.pk)

        return {
            'include_author_name': include_author_name,
            'to_expand': list(to_expand),
        }


class MainFieldSet(BaseReviewRequestFieldSet):
    fieldset_id = 'main'
    field_classes = [
        SummaryField,
        DescriptionField,
        TestingDoneField,
    ]


class ExtraFieldSet(BaseReviewRequestFieldSet):
    """A field set that is displayed after the main field set."""

    fieldset_id = 'extra'
    field_classes = [
        CommitListField,
    ]


class InformationFieldSet(BaseReviewRequestFieldSet):
    fieldset_id = 'info'
    label = _('Information')
    field_classes = [
        OwnerField,
        RepositoryField,
        BranchField,
        BugsField,
        DependsOnField,
        BlocksField,
        ChangeField,
        CommitField,
    ]


class ReviewersFieldSet(BaseReviewRequestFieldSet):
    fieldset_id = 'reviewers'
    label = _('Reviewers')
    show_required = True
    field_classes = [
        TargetGroupsField,
        TargetPeopleField,
    ]


class ChangeEntryOnlyFieldSet(BaseReviewRequestFieldSet):
    fieldset_id = '_change_entries_only'
    field_classes = [
        DiffField,
        FileAttachmentCaptionsField,
        ScreenshotCaptionsField,
        FileAttachmentsField,
        ScreenshotsField,
        StatusField,
    ]


builtin_fieldsets = [
    MainFieldSet,
    ExtraFieldSet,
    InformationFieldSet,
    ReviewersFieldSet,
    ChangeEntryOnlyFieldSet,
]



"""Add Group.email_list_only.

Version Added:
    2.5
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('Group', 'email_list_only', models.BooleanField, initial=True)
]



"""Unit tests for viewing drafts as other users.

Version Added:
    7.0.2
"""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING, Union, cast

from django.contrib.auth.models import Permission, User

from reviewboard.reviews.context import ReviewRequestContext
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from reviewboard.reviews.models import ReviewRequest, ReviewRequestDraft


class OtherUserDraftTests(TestCase):
    """Unit tests for viewing drafts as other users.

    Admins or users with the reviews.can_edit_reviewrequest permission are
    allowed to see and manipulate drafts on review requests owned by other
    people. These tests cover the various reviewable pages and check that we're
    adding the right things to the page data.

    Version Added:
        7.0.2
    """

    fixtures = ['test_scmtools', 'test_users']

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        users = User.objects.in_bulk(
            ['doc', 'grumpy', 'admin', 'dopey'],
            field_name='username')
        self.user_owner = users['doc']
        self.user_regular = users['grumpy']
        self.user_admin = users['admin']
        self.user_privileged = users['dopey']
        self.user_privileged.user_permissions.add(
            Permission.objects.get(codename='can_edit_reviewrequest'))

    def test_unpublished_review_request_as_owner(self) -> None:
        """Testing ReviewRequestDetailView with an unpublished review request
        as the owner
        """
        review_request = self.create_review_request()
        draft = self.create_review_request_draft(review_request)

        self.client.force_login(self.user_owner)
        rsp = self.client.get(review_request.get_absolute_url())

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_unpublished_review_request_as_regular_user(self) -> None:
        """Testing ReviewRequestDetailView with an unpublished review request
        as a regular user
        """
        review_request = self.create_review_request()
        self.create_review_request_draft(review_request)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(review_request.get_absolute_url())

        self.assertEqual(rsp.status_code, 403)

    def test_unpublished_review_request_as_admin(self) -> None:
        """Testing ReviewRequestDetailView with an unpublished review request
        as an admin
        """
        review_request = self.create_review_request()
        draft = self.create_review_request_draft(review_request)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(review_request.get_absolute_url())

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_unpublished_review_request_as_privileged(self) -> None:
        """Testing ReviewRequestDetailView with an unpublished review request
        as a privileged user
        """
        review_request = self.create_review_request()
        draft = self.create_review_request_draft(review_request)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(review_request.get_absolute_url())

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_review_request_draft_as_owner(self) -> None:
        """Testing ReviewRequestDetailView with a draft on a published review
        request as the owner
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)

        self.client.force_login(self.user_owner)
        rsp = self.client.get(review_request.get_absolute_url())

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_review_request_draft_as_regular_user(self) -> None:
        """Testing ReviewRequestDetailView with a draft on a published review
        request as a regular user
        """
        review_request = self.create_review_request(publish=True)
        self.create_review_request_draft(review_request)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(review_request.get_absolute_url())

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=None,
            review_request_details=review_request,
            mutable_by_user=False,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_review_request_draft_as_admin(self) -> None:
        """Testing ReviewRequestDetailView with a draft on a published review
        request as an admin
        """
        review_request = self.create_review_request(publish=True)
        self.create_review_request_draft(review_request)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(review_request.get_absolute_url())

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=None,
            review_request_details=review_request,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=False)

    def test_review_request_draft_as_privileged(self) -> None:
        """Testing ReviewRequestDetailView with a draft on a published review
        request as a privileged user
        """
        review_request = self.create_review_request(publish=True)
        self.create_review_request_draft(review_request)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(review_request.get_absolute_url())

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=None,
            review_request_details=review_request,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=False)

    def test_review_request_draft_as_owner_with_view_draft(self) -> None:
        """Testing ReviewRequestDetailView with a draft on a published review
        request as the owner with ?view-draft=1
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)

        self.client.force_login(self.user_owner)
        rsp = self.client.get(review_request.get_absolute_url() +
                              '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_review_request_draft_as_regular_user_with_view_draft(
        self,
    ) -> None:
        """Testing ReviewRequestDetailView with a draft on a published review
        request as a regular user with ?view-draft=1
        """
        review_request = self.create_review_request(publish=True)
        self.create_review_request_draft(review_request)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(review_request.get_absolute_url() +
                              '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=None,
            review_request_details=review_request,
            mutable_by_user=False,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_review_request_draft_as_admin_with_view_draft(self) -> None:
        """Testing ReviewRequestDetailView with a draft on a published review
        request as an admin with ?view-draft=1
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(review_request.get_absolute_url() +
                              '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_review_request_draft_as_privileged_user_with_view_draft(
        self,
    ) -> None:
        """Testing ReviewRequestDetailView with a draft on a published review
        request as a privileged_user with ?view-draft=1
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(review_request.get_absolute_url() +
                              '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_unpublished_diff_as_owner(self) -> None:
        """Testing ReviewsDiffViewerView with an unpublished review request as
        the owner
        """
        review_request = self.create_review_request(create_repository=True)
        diffset = self.create_diffset(review_request, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_unpublished_diff_as_regular_user(self) -> None:
        """Testing ReviewsDiffViewerView with an unpublished review request as
        a regular user
        """
        review_request = self.create_review_request(create_repository=True)
        diffset = self.create_diffset(review_request, draft=True)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }))

        self.assertEqual(rsp.status_code, 403)

    def test_unpublished_diff_as_admin(self) -> None:
        """Testing ReviewsDiffViewerView with an unpublished review request as
        an admin
        """
        review_request = self.create_review_request(create_repository=True)
        diffset = self.create_diffset(review_request, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_unpublished_diff_as_privileged(self) -> None:
        """Testing ReviewsDiffViewerView with an unpublished review request as
        a privileged user
        """
        review_request = self.create_review_request(create_repository=True)
        diffset = self.create_diffset(review_request, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_draft_diff_as_owner(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as the owner"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        self.create_diffset(review_request)
        diffset = self.create_diffset(review_request, revision=2, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_draft_diff_as_regular_user(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as a regular
        user
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        self.create_diffset(review_request)
        diffset = self.create_diffset(review_request, revision=2, draft=True)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }))

        self.assertEqual(rsp.status_code, 404)

    def test_draft_diff_as_admin(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as an admin"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        self.create_diffset(review_request)
        diffset = self.create_diffset(review_request, revision=2, draft=True)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=None,
            review_request_details=review_request,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=False)

    def test_draft_diff_as_privileged(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as a privileged
        user
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        self.create_diffset(review_request)
        diffset = self.create_diffset(review_request, revision=2, draft=True)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=None,
            review_request_details=review_request,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=False)

    def test_draft_diff_as_owner_with_view_draft(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as the owner
        with ?view-draft=1
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        self.create_diffset(review_request)
        diffset = self.create_diffset(review_request, revision=2, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_draft_diff_as_regular_user_with_view_draft(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as a regular
        user with ?view-draft=1
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        self.create_diffset(review_request)
        diffset = self.create_diffset(review_request, revision=2, draft=True)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 404)

    def test_draft_diff_as_admin_with_view_draft(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as an admin with
        ?view-draft=1
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        self.create_diffset(review_request)
        diffset = self.create_diffset(review_request, revision=2, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_draft_diff_as_privileged_with_view_draft(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as a privileged
        user with ?view-draft=1
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        self.create_diffset(review_request)
        diffset = self.create_diffset(review_request, revision=2, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_draft_interdiff_as_owner(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as the owner"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset1 = self.create_diffset(review_request)
        diffset2 = self.create_diffset(review_request, revision=2, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'view-interdiff',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset1.revision,
                    'interdiff_revision': diffset2.revision,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_draft_interdiff_as_regular_user(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as a regular
        user
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset1 = self.create_diffset(review_request)
        diffset2 = self.create_diffset(review_request, revision=2, draft=True)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'view-interdiff',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset1.revision,
                    'interdiff_revision': diffset2.revision,
                }))

        self.assertEqual(rsp.status_code, 404)

    def test_draft_interdiff_as_admin(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as an admin"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset1 = self.create_diffset(review_request)
        diffset2 = self.create_diffset(review_request, revision=2, draft=True)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'view-interdiff',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset1.revision,
                    'interdiff_revision': diffset2.revision,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=None,
            review_request_details=review_request,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=False)

    def test_draft_interdiff_as_privileged(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as a privileged
        user
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset1 = self.create_diffset(review_request)
        diffset2 = self.create_diffset(review_request, revision=2, draft=True)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'view-interdiff',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset1.revision,
                    'interdiff_revision': diffset2.revision,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=None,
            review_request_details=review_request,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=False)

    def test_draft_interdiff_as_owner_with_view_draft(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as the owner
        with ?view-draft=1
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset1 = self.create_diffset(review_request)
        diffset2 = self.create_diffset(review_request, revision=2, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'view-interdiff',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset1.revision,
                    'interdiff_revision': diffset2.revision,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_draft_interdiff_as_regular_user_with_view_draft(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as a regular
        user with ?view-draft=1
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset1 = self.create_diffset(review_request)
        diffset2 = self.create_diffset(review_request, revision=2, draft=True)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'view-interdiff',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset1.revision,
                    'interdiff_revision': diffset2.revision,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 404)

    def test_draft_interdiff_as_admin_with_view_draft(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as an admin with
        ?view-draft=1
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset1 = self.create_diffset(review_request)
        diffset2 = self.create_diffset(review_request, revision=2, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'view-interdiff',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset1.revision,
                    'interdiff_revision': diffset2.revision,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_draft_interdiff_as_privileged_with_view_draft(self) -> None:
        """Testing ReviewsDiffViewerView with a new draft diff as a privileged
        user with ?view-draft=1
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset1 = self.create_diffset(review_request)
        diffset2 = self.create_diffset(review_request, revision=2, draft=True)
        draft = review_request.draft.get()

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'view-interdiff',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset1.revision,
                    'interdiff_revision': diffset2.revision,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_unpublished_file_attachment_as_owner(self) -> None:
        """Testing ReviewFileAttachmentView with an unpublished review request
        as the owner
        """
        review_request = self.create_review_request()
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_unpublished_file_attachment_as_regular_user(self) -> None:
        """Testing ReviewFileAttachmentView with an unpublished review request
        as a regular user
        """
        review_request = self.create_review_request()
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }))

        self.assertEqual(rsp.status_code, 403)

    def test_unpublished_file_attachment_as_admin(self) -> None:
        """Testing ReviewFileAttachmentView with an unpublished review request
        as an admin
        """
        review_request = self.create_review_request()
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_unpublished_file_attachment_as_privileged(self) -> None:
        """Testing ReviewFileAttachmentView with an unpublished review request
        as a privileged user
        """
        review_request = self.create_review_request()
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_draft_file_attachment_as_owner(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as the
        owner
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_draft_file_attachment_as_regular_user(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as a
        regular user
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }))

        self.assertEqual(rsp.status_code, 404)

    def test_draft_file_attachment_as_admin(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as an
        admin
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }))

        self.assertEqual(rsp.status_code, 404)

    def test_draft_file_attachment_as_privileged(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as a
        privileged user
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }))

        self.assertEqual(rsp.status_code, 404)

    def test_draft_file_attachment_diff_as_owner(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as the
        owner
        """
        review_request = self.create_review_request(publish=True)
        attachment1 = self.create_file_attachment(review_request)
        draft = self.create_review_request_draft(review_request)
        attachment2 = self.create_file_attachment(
            review_request, draft=draft,
            attachment_history=attachment1.attachment_history)

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_diff_id': attachment1.pk,
                    'file_attachment_id': attachment2.pk,
                }))

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_draft_file_attachment_diff_as_regular_user(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as a
        regular user
        """
        review_request = self.create_review_request(publish=True)
        attachment1 = self.create_file_attachment(review_request)
        draft = self.create_review_request_draft(review_request)
        attachment2 = self.create_file_attachment(
            review_request, draft=draft,
            attachment_history=attachment1.attachment_history)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_diff_id': attachment1.pk,
                    'file_attachment_id': attachment2.pk,
                }))

        self.assertEqual(rsp.status_code, 404)

    def test_draft_file_attachment_diff_as_admin(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as an
        admin
        """
        review_request = self.create_review_request(publish=True)
        attachment1 = self.create_file_attachment(review_request)
        draft = self.create_review_request_draft(review_request)
        attachment2 = self.create_file_attachment(
            review_request, draft=draft,
            attachment_history=attachment1.attachment_history)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_diff_id': attachment1.pk,
                    'file_attachment_id': attachment2.pk,
                }))

        self.assertEqual(rsp.status_code, 404)

    def test_draft_file_attachment_diff_as_privileged(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as a
        privileged user
        """
        review_request = self.create_review_request(publish=True)
        attachment1 = self.create_file_attachment(review_request)
        draft = self.create_review_request_draft(review_request)
        attachment2 = self.create_file_attachment(
            review_request, draft=draft,
            attachment_history=attachment1.attachment_history)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_diff_id': attachment1.pk,
                    'file_attachment_id': attachment2.pk,
                }))

        self.assertEqual(rsp.status_code, 404)

    def test_draft_file_attachment_as_owner_with_view_draft(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as the
        owner
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_draft_file_attachment_as_regular_user_with_view_draft(
        self,
    ) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as a
        regular user
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 404)

    def test_draft_file_attachment_as_admin_with_view_draft(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as an
        admin
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_draft_file_attachment_as_privileged_with_view_draft(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as a
        privileged user
        """
        review_request = self.create_review_request(publish=True)
        draft = self.create_review_request_draft(review_request)
        attachment = self.create_file_attachment(review_request, draft=draft)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_id': attachment.pk,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_draft_file_attachment_diff_as_owner_with_view_draft(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as the
        owner
        """
        review_request = self.create_review_request(publish=True)
        attachment1 = self.create_file_attachment(review_request)
        draft = self.create_review_request_draft(review_request)
        attachment2 = self.create_file_attachment(
            review_request, draft=draft,
            attachment_history=attachment1.attachment_history)

        self.client.force_login(self.user_owner)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_diff_id': attachment1.pk,
                    'file_attachment_id': attachment2.pk,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=False,
            user_draft_exists=False,
            viewing_user_draft=False)

    def test_draft_file_attachment_diff_as_regular_user_with_view_draft(
        self,
    ) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as a
        regular user
        """
        review_request = self.create_review_request(publish=True)
        attachment1 = self.create_file_attachment(review_request)
        draft = self.create_review_request_draft(review_request)
        attachment2 = self.create_file_attachment(
            review_request, draft=draft,
            attachment_history=attachment1.attachment_history)

        self.client.force_login(self.user_regular)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_diff_id': attachment1.pk,
                    'file_attachment_id': attachment2.pk,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 404)

    def test_draft_file_attachment_diff_as_admin_with_view_draft(self) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as an
        admin
        """
        review_request = self.create_review_request(publish=True)
        attachment1 = self.create_file_attachment(review_request)
        draft = self.create_review_request_draft(review_request)
        attachment2 = self.create_file_attachment(
            review_request, draft=draft,
            attachment_history=attachment1.attachment_history)

        self.client.force_login(self.user_admin)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_diff_id': attachment1.pk,
                    'file_attachment_id': attachment2.pk,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def test_draft_file_attachment_diff_as_privileged_with_view_draft(
        self,
    ) -> None:
        """Testing ReviewFileAttachmentView with a draft attachment as a
        privileged user
        """
        review_request = self.create_review_request(publish=True)
        attachment1 = self.create_file_attachment(review_request)
        draft = self.create_review_request_draft(review_request)
        attachment2 = self.create_file_attachment(
            review_request, draft=draft,
            attachment_history=attachment1.attachment_history)

        self.client.force_login(self.user_privileged)
        rsp = self.client.get(
            local_site_reverse(
                'file-attachment',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'file_attachment_diff_id': attachment1.pk,
                    'file_attachment_id': attachment2.pk,
                }) + '?view-draft=1')

        self.assertEqual(rsp.status_code, 200)
        self._test_result_context(
            context=cast(ReviewRequestContext, rsp.context),
            draft=draft,
            review_request_details=draft,
            mutable_by_user=True,
            force_view_user_draft=True,
            user_draft_exists=True,
            viewing_user_draft=True)

    def _test_result_context(
        self,
        *,
        context: ReviewRequestContext,
        draft: Optional[ReviewRequestDraft],
        review_request_details: Union[ReviewRequest, ReviewRequestDraft],
        mutable_by_user: bool,
        force_view_user_draft: bool,
        user_draft_exists: bool,
        viewing_user_draft: bool,
    ) -> None:
        """Do assertions for the result context.

        Args:
            context (reviewboard.reviews.context.ReviewRequestContext):
                The rendering context for the review request page.

            draft (reviewboard.reviews.models.ReviewRequestDraft):
                The draft that should be set in the context. This may be None.

            review_request_details (reviewboard.reviews.models.ReviewRequest or
                                    reviewboard.reviews.models.
                                    ReviewRequestDraft):
                The object that should be used for showing the review request
                data.

            mutable_by_user (bool):
                Whether the review request should be mutable by the current
                user.

            force_view_user_draft (bool):
                Whether the user is forced to view a draft that is owned by
                another user (such as an admin viewing an unpublished review
                request).

            user_draft_exists (bool):
                Whether a draft exists that is owned by another user.

            viewing_user_draft (bool):
                Whether the user is viewing a draft that is owned by another
                user.
        """
        self.assertEqual(context['draft'], draft)
        self.assertEqual(context['review_request_details'],
                         review_request_details)
        self.assertEqual(context['mutable_by_user'],
                         mutable_by_user)
        self.assertEqual(context['force_view_user_draft'],
                         force_view_user_draft)
        self.assertEqual(context['user_draft_exists'],
                         user_draft_exists)
        self.assertEqual(context['viewing_user_draft'],
                         viewing_user_draft)



"""Notification-related forms."""

from __future__ import annotations

from typing import Any, Optional

from django import forms
from django.core.validators import URLValidator
from django.forms.fields import CharField, MultipleChoiceField
from django.utils.translation import gettext_lazy as _

from reviewboard.admin.form_widgets import RelatedRepositoryWidget
from reviewboard.notifications.models import WebHookTarget
from reviewboard.scmtools.models import Repository
from reviewboard.site.mixins import LocalSiteAwareModelFormMixin


class WebHookTargetForm(LocalSiteAwareModelFormMixin, forms.ModelForm):
    """A form for creating and updating WebHookTargets."""

    url = CharField(
        label=_('URL'),
        validators=[URLValidator()],
        widget=forms.widgets.URLInput(attrs={'size': 100}),
    )

    events = MultipleChoiceField(
        choices=WebHookTarget.EVENT_CHOICES,
        required=False,
        widget=forms.widgets.CheckboxSelectMultiple)

    repositories = forms.ModelMultipleChoiceField(
        label=_('Repositories'),
        required=False,
        queryset=Repository.objects.filter(visible=True).order_by('name'),
        widget=RelatedRepositoryWidget())

    def clean_extra_data(self) -> Optional[str]:
        """Ensure that extra_data is a valid value.

        Returns:
            str:
            Either a non-zero length string of JSON-encoded extra data or None.
        """
        return self.cleaned_data['extra_data'] or None

    def clean_events(self) -> list[str]:
        """Clean the "events" field.

        Returns:
            list of str:
            The cleaned events field data.
        """
        events = self.cleaned_data['events']

        if '*' in events:
            events = ['*']

        return events

    def clean(self) -> dict[str, Any]:
        """Validate the state of the entire form.

        Returns:
            The cleaned form data.
        """
        super().clean()

        custom_content = self.cleaned_data.get('custom_content', '')
        self.cleaned_data['use_custom_content'] = len(custom_content) > 0

        apply_to = self.cleaned_data.get('apply_to')

        if (apply_to != WebHookTarget.APPLY_TO_SELECTED_REPOS or
            'repositories' not in self.cleaned_data):
            self.cleaned_data['repositories'] = Repository.objects.none()

        return self.cleaned_data

    class Meta:
        """Metadata for the WebHookTarget form."""

        model = WebHookTarget
        widgets = {
            'apply_to': forms.widgets.RadioSelect(),
        }
        error_messages = {
            'repositories': {
                'invalid_choice': _('A repository with ID %(value)s was not '
                                    'found.'),
                'invalid_pk_value': _('"%(pk)s" is an invalid repository ID.'),
            },
        }
        fields = '__all__'



"""Root view for reviews."""

from django.views.generic.base import RedirectView

from reviewboard.accounts.mixins import (CheckLoginRequiredViewMixin,
                                         UserProfileRequiredViewMixin)
from reviewboard.site.mixins import CheckLocalSiteAccessViewMixin
from reviewboard.site.urlresolvers import local_site_reverse


class RootView(CheckLoginRequiredViewMixin,
               UserProfileRequiredViewMixin,
               CheckLocalSiteAccessViewMixin,
               RedirectView):
    """Handles the root URL of Review Board or a Local Site.

    If the user is authenticated, this will redirect to their Dashboard.
    Otherwise, they'll be redirected to the All Review Requests page.

    Either page may then redirect for login or show a Permission Denied,
    depending on the settings.
    """

    permanent = False

    def get_redirect_url(
        self,
        *args,
        **kwargs,
    ) -> str:
        """Return the URL to redirect to.

        Args:
            *args (tuple):
                Positional arguments passed to the view.

            **kwargs (dict):
                Keyword arguments passed to the view.

        Returns:
            str:
            The URL to redirect to. If the user is authenticated, this will
            return the dashboard's URL. Otherwise, it will return the
            All Review Request page's URL.
        """
        if self.request.user.is_authenticated:
            url_name = 'dashboard'
        else:
            url_name = 'all-review-requests'

        return local_site_reverse(url_name, local_site=self.local_site)



"""The DiffCommit validation resource."""

import logging

from django.db.models import Q
from djblets.webapi.decorators import (webapi_login_required,
                                       webapi_request_fields,
                                       webapi_response_errors)
from djblets.webapi.errors import (DOES_NOT_EXIST, INVALID_ATTRIBUTE,
                                   INVALID_FORM_DATA, NOT_LOGGED_IN,
                                   PERMISSION_DENIED)
from djblets.webapi.fields import FileFieldType, StringFieldType

from reviewboard.certs.errors import CertificateVerificationError
from reviewboard.diffviewer.commit_utils import (serialize_validation_info,
                                                 update_validation_info)
from reviewboard.diffviewer.errors import (DiffParserError,
                                           DiffTooBigError,
                                           EmptyDiffError)
from reviewboard.diffviewer.features import dvcs_feature
from reviewboard.diffviewer.forms import ValidateCommitForm
from reviewboard.scmtools.errors import FileNotFoundError, SCMError
from reviewboard.scmtools.git import ShortSHA1Error
from reviewboard.scmtools.models import Repository
from reviewboard.webapi.base import WebAPIResource
from reviewboard.webapi.decorators import (webapi_check_local_site,
                                           webapi_check_login_required)
from reviewboard.webapi.errors import (DIFF_EMPTY,
                                       DIFF_PARSE_ERROR,
                                       DIFF_TOO_BIG,
                                       INVALID_REPOSITORY,
                                       REPO_FILE_NOT_FOUND,
                                       UNVERIFIED_HOST_CERT)


logger = logging.getLogger(__name__)


class ValidateDiffCommitResource(WebAPIResource):
    """Verifies whether or not a diff file for a commit will work.

    This allows clients to validate whether or not diff files for commits can
    be parsed and displayed without actually creating a review request first.
    """

    added_in = '4.0'

    singleton = True
    name = 'commit_validation'
    uri_name = 'commits'
    uri_object_key = None
    model = None

    allowed_methods = ('GET', 'POST')

    required_features = [dvcs_feature]

    item_child_resources = []
    list_child_resources = []

    fields = {
        'validation_info': {
            'type': StringFieldType,
            'description': (
                'Validation metadata to pass to this resource to help '
                'validate the next commit.'
            ),
        },
    }

    @webapi_check_local_site
    @webapi_check_login_required
    def get(self, request, *args, **kwargs):
        """Return links for using this resource."""
        return 200, {
            'links': self.get_links(request=request, *args, **kwargs),
        }

    @webapi_login_required
    @webapi_check_local_site
    @webapi_response_errors(
        DIFF_EMPTY,
        DIFF_PARSE_ERROR,
        DIFF_TOO_BIG,
        DOES_NOT_EXIST,
        INVALID_ATTRIBUTE,
        INVALID_FORM_DATA,
        INVALID_REPOSITORY,
        NOT_LOGGED_IN,
        REPO_FILE_NOT_FOUND,
        PERMISSION_DENIED,
        UNVERIFIED_HOST_CERT,
    )
    @webapi_request_fields(
        required={
            'repository': {
                'type': StringFieldType,
                'description': 'The path or ID of the repository.',
            },
            'diff': {
                'type': FileFieldType,
                'description': 'The diff file to validate.',
            },
            'commit_id': {
                'type': StringFieldType,
                'description': 'The ID of the commit being validated.',
            },
            'parent_id': {
                'type': StringFieldType,
                'description': 'The ID of the parent commit.',
            },
        },
        optional={
            'base_commit_id': {
                'type': StringFieldType,
                'description': 'The base commit ID.',
            },
            'parent_diff': {
                'type': FileFieldType,
                'description': (
                    'The parent diff of the commit being validated.\n'
                ),
            },
            'validation_info': {
                'type': StringFieldType,
                'description': (
                    'Validation metadata from a previous call to this API.\n'
                    '\n'
                    'This field is required for all but the first commit in a '
                    'series.'
                ),
            },
        }
    )
    def create(self, request, repository, commit_id, parent_id,
               base_commit_id=None, local_site_name=None, *args, **kwargs):
        """Validate a diff for a commit.

        This API has a similar signature to the :ref:`Draft DiffCommit resource
        <webapi2.0-draft-diff-commit-list-resource>` POST API, but instead of
        actually creating commits, it will return a result representing whether
        or not the included diff file parsed and validated correctly.

        This API must be called before posting to the :ref:`Draft DiffCommit
        resource <webapi2.0-draft-diff-commit-list-resource>` because the
        ``validation_info`` field returned by this resource is required for
        posting to that resource.
        """
        local_site = self._get_local_site(local_site_name)

        try:
            q = Q(pk=int(repository))
        except ValueError:
            q = (Q(path=repository) |
                 Q(mirror_path=repository) |
                 Q(name=repository))

        repository_qs = (
            Repository.objects
            .accessible(request.user, local_site=local_site)
            .filter(q)
        )
        repository_count = len(repository_qs)

        if repository_count == 0:
            return INVALID_REPOSITORY, {
                'repository': repository,
            }
        elif repository_count > 1:
            msg = (
                'Too many repositories matched "%s". Try specifying the '
                'repository by name instead.'
                % repository
            )

            return INVALID_REPOSITORY.with_message(msg), {
                'repository': repository,
            }

        repository = repository_qs.first()

        if not repository.scmtool_class.supports_history:
            return INVALID_ATTRIBUTE, {
                'reason': (
                    'The "%s" repository does not support review requests '
                    'created with history.'
                    % repository.name
                ),
            }

        form = ValidateCommitForm(repository=repository,
                                  request=request,
                                  data=request.POST,
                                  files=request.FILES)

        if not form.is_valid():
            return INVALID_FORM_DATA, {
                'fields': self._get_form_errors(form),
            }

        try:
            filediffs = form.validate_diff()
        except FileNotFoundError as e:
            return REPO_FILE_NOT_FOUND, {
                'file': e.path,
                'revision': str(e.revision),
            }
        except EmptyDiffError:
            return DIFF_EMPTY
        except DiffTooBigError as e:
            return DIFF_TOO_BIG, {
                'reason': str(e),
                'max_size': e.max_diff_size,
            }
        except DiffParserError as e:
            return DIFF_PARSE_ERROR, {
                'reason': str(e),
                'linenum': e.linenum,
            }
        except ShortSHA1Error as e:
            return REPO_FILE_NOT_FOUND, {
                'reason': str(e),
                'file': e.path,
                'revision': str(e.revision),
            }
        except CertificateVerificationError as e:
            return UNVERIFIED_HOST_CERT.with_message(str(e)), {
                'certificate': e.certificate,
            }
        except SCMError as e:
            return DIFF_PARSE_ERROR.with_message(str(e))
        except Exception as e:
            logger.exception(
                'Unexpected exception occurred while validating commit "%s" '
                'in repository "%s" (id %d) with base_commit_id="%s"',
                commit_id,
                repository.name,
                repository.pk,
                base_commit_id,
                extra={'request': request})
            return DIFF_PARSE_ERROR.with_message(
                'Unexpected error while validating the diff: %s' % e)

        validation_info = update_validation_info(
            form.cleaned_data.get('validation_info', {}),
            commit_id,
            parent_id,
            filediffs)

        return 200, {
            self.item_result_key: {
                'validation_info': serialize_validation_info(
                    validation_info),
            }
        }


validate_diffcommit_resource = ValidateDiffCommitResource()



"""Unit tests for reviewboard.accounts.models.Trophy."""

from django.contrib.auth.models import User

from reviewboard.accounts.models import Trophy
from reviewboard.testing import TestCase


class TrophyTests(TestCase):
    """Unit tests for reviewboard.accounts.models.Trophy."""

    fixtures = ['test_users']

    def test_is_fish_trophy_awarded_for_new_review_request(self):
        """Testing if a fish trophy is awarded for a new review request"""
        user1 = User.objects.get(username='doc')
        category = 'fish'
        review_request = self.create_review_request(publish=True, id=3223,
                                                    submitter=user1)
        trophies = Trophy.objects.get_trophies(review_request)
        self.assertEqual(trophies[0].category, category)
        self.assertTrue(
            trophies[0].review_request.extra_data['calculated_trophies'])

    def test_is_fish_trophy_awarded_for_older_review_request(self):
        """Testing if a fish trophy is awarded for an older review request"""
        user1 = User.objects.get(username='doc')
        category = 'fish'
        review_request = self.create_review_request(publish=True, id=1001,
                                                    submitter=user1)
        del review_request.extra_data['calculated_trophies']
        trophies = Trophy.objects.get_trophies(review_request)
        self.assertEqual(trophies[0].category, category)
        self.assertTrue(
            trophies[0].review_request.extra_data['calculated_trophies'])

    def test_is_milestone_trophy_awarded_for_new_review_request(self):
        """Testing if a milestone trophy is awarded for a new review request
        """
        user1 = User.objects.get(username='doc')
        category = 'milestone'
        review_request = self.create_review_request(publish=True, id=1000,
                                                    submitter=user1)
        trophies = Trophy.objects.compute_trophies(review_request)
        self.assertEqual(trophies[0].category, category)
        self.assertTrue(
            trophies[0].review_request.extra_data['calculated_trophies'])

    def test_is_milestone_trophy_awarded_for_older_review_request(self):
        """Testing if a milestone trophy is awarded for an older review
        request
        """
        user1 = User.objects.get(username='doc')
        category = 'milestone'
        review_request = self.create_review_request(publish=True, id=10000,
                                                    submitter=user1)
        del review_request.extra_data['calculated_trophies']
        trophies = Trophy.objects.compute_trophies(review_request)
        self.assertEqual(trophies[0].category, category)
        self.assertTrue(
            trophies[0].review_request.extra_data['calculated_trophies'])

    def test_is_no_trophy_awarded(self):
        """Testing if no trophy is awarded"""
        user1 = User.objects.get(username='doc')
        review_request = self.create_review_request(publish=True, id=999,
                                                    submitter=user1)
        trophies = Trophy.objects.compute_trophies(review_request)
        self.assertFalse(trophies)



"""Increase Repository.name's max length to 255 characters.

The old max length was 64 characters.

Version Added:
    3.0.11
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('Repository', 'name', initial=None, max_length=255),
]



"""Registries for Review Board UI theme support.

Version Added:
    7.0
"""

from __future__ import annotations

from typing import Iterator, Optional, cast

from djblets.siteconfig.models import SiteConfiguration

from reviewboard.registries.registry import OrderedRegistry
from reviewboard.themes.ui.base import BaseUITheme


class UIThemeRegistry(OrderedRegistry[BaseUITheme]):
    """Registry for managing available Review Board UI themes.

    This provides functionality for looking up available UI themes,
    determining a default theme, and registering new themes.

    Version Changed:
        8.0:
        The ``system`` theme is now the default when using the ``default``
        theme.

    Version Added:
        7.0
    """

    DEFAULT_THEME_ID = 'system'

    lookup_attrs = ['theme_id']

    def get_theme(
        self,
        theme_id: str,
    ) -> Optional[BaseUITheme]:
        """Return a theme with the specified ID.

        A special value of "default" will always return the default theme
        ID.

        Args:
            theme_id (str):
                The ID of the theme to return.

        Returns:
            reviewboard.themes.ui.base.BaseUITheme:
            The theme instance, or ``None`` if not found.
        """
        if theme_id == 'default':
            theme_id = self.get_default_theme_id()

        return self.get('theme_id', theme_id)

    def get_default_theme_id(self) -> str:
        """Return the ID of the default theme for the UI.

        If the site configuration contains a ``default_ui_theme`` value,
        then that theme ID will be returned, if found in the registry.

        This will always fall back to :py:attr:`DEFAULT_THEME_ID` if a
        configured value is not found.

        Returns:
            str:
            The default theme ID.
        """
        siteconfig = SiteConfiguration.objects.get_current()
        default_theme_id = cast(str, siteconfig.get('default_ui_theme'))

        if (default_theme_id and
            self.get('theme_id', default_theme_id) is not None):
            # The configured default is valid.
            return default_theme_id

        return self.DEFAULT_THEME_ID

    def get_defaults(self) -> Iterator[BaseUITheme]:
        """Return defaults for the registry.

        Yields:
            type:
            Each default UI theme.
        """
        from reviewboard.themes.ui.default import (
            DarkUITheme,
            LightUITheme,
            SystemUITheme,
        )

        yield from (
            LightUITheme(),
            DarkUITheme(),
            SystemUITheme(),
        )


#: The main registry for UI themes.
#:
#: Version Added:
#:     7.0
ui_theme_registry = UIThemeRegistry()
